import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock, MagicMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock, MagicMock

from pathlib import Path
from unittest.mock import Mock, call

import pytest

from vedro import Scenario
from vedro.core import ScenarioDiscoverer, ScenarioFinder, ScenarioLoader, VirtualScenario


@pytest.fixture()
def finder_factory():
    def _factory(files):
        iterator = MagicMock()
        iterator.__aiter__.return_value = iter(files)
        return Mock(ScenarioFinder, find=Mock(return_value=iterator))
    return _factory


@pytest.fixture()
def loader_factory():
    def _factory(scenarios):
        return Mock(ScenarioLoader, load=AsyncMock(side_effect=iter(scenarios)))
    return _factory


def create_scenario(filename):
    return Mock(Scenario, __file__=filename)


@pytest.mark.asyncio
async def test_scenario_discoverer(*, finder_factory, loader_factory):
    root = Path("/tmp")
    scenario1 = create_scenario(root / "scenario-1.py")
    scenario3 = create_scenario(root / "folder" / "scenario-2.py")
    scenario4 = create_scenario(root / "folder" / "scenario-10.py")
    tree = {
        scenario1.__file__: [scenario1],
        root / "scenario-2.py": [],
        scenario3.__file__: [scenario3, scenario4],
    }
    finder_ = finder_factory(tree.keys())
    loader_ = loader_factory(tree.values())
    discoverer = ScenarioDiscoverer(finder_, loader_)

    scenarios = await discoverer.discover(root)
    assert scenarios == [
        VirtualScenario(scenario1, []),
        VirtualScenario(scenario3, []),
        VirtualScenario(scenario4, []),
    ]
    assert finder_.mock_calls == [call.find(root)]
    assert loader_.mock_calls == [call.load(f) for f in tree.keys()]



from unittest.mock import Mock

import pytest
from baby_steps import given, then, when

from vedro.core import Dispatcher
from vedro.events import ScenarioSkippedEvent
from vedro.plugins.director import DirectorPlugin, RichReporterPlugin
from vedro.plugins.director.rich.test_utils import (
    chose_reporter,
    console_,
    director,
    dispatcher,
    make_scenario_result,
    reporter,
)

__all__ = ("dispatcher", "reporter", "director", "console_",)


@pytest.mark.asyncio
async def test_rich_reporter_scenario_skip_event(*, dispatcher: Dispatcher,
                                                 director: DirectorPlugin,
                                                 reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        scenario_result = make_scenario_result()
        event = ScenarioSkippedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []



import sys
from typing import Callable, Type

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import CleanupEvent

__all__ = ("Terminator", "TerminatorPlugin",)


class TerminatorPlugin(Plugin):
    def __init__(self, config: Type["Terminator"], *, exit_fn: Callable[[int], None] = sys.exit):
        super().__init__(config)
        self._exit_fn = exit_fn

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(CleanupEvent, self.on_cleanup, priority=sys.maxsize)

    def on_cleanup(self, event: CleanupEvent) -> None:
        if event.report.failed > 0 or event.report.passed == 0:
            self._exit_fn(1)
        else:
            self._exit_fn(0)


class Terminator(PluginConfig):
    plugin = TerminatorPlugin



from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when
from rich.console import Style

from vedro.core import Dispatcher, Report
from vedro.events import CleanupEvent
from vedro.plugins.director import DirectorPlugin, RichReporterPlugin
from vedro.plugins.director.rich.test_utils import (
    chose_reporter,
    console_,
    director,
    dispatcher,
    make_report,
    make_scenario_result,
    reporter,
)

__all__ = ("dispatcher", "reporter", "director", "console_",)


@pytest.mark.asyncio
async def test_rich_reporter_cleanup_event(*, dispatcher: Dispatcher, director: DirectorPlugin,
                                           reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        report = Report()
        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(" "),
            call.out("# 0 scenarios, 0 passed, 0 failed, 0 skipped",
                     style=Style.parse("bold red"),
                     end=""),
            call.out(f" ({report.elapsed:.2f}s)", style=Style.parse("blue"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_cleanup_event_with_summary(*, dispatcher: Dispatcher,
                                                        director: DirectorPlugin,
                                                        reporter: RichReporterPlugin,
                                                        console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        summaries = ["<summary1>", "<summary2>"]
        report = make_report(summaries=summaries)
        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(" "),
            call.out("# " + "\n# ".join(summaries), style=Style.parse("grey70")),
            call.out("# 0 scenarios, 0 passed, 0 failed, 0 skipped",
                     style=Style.parse("bold red"),
                     end=""),
            call.out(f" ({report.elapsed:.2f}s)", style=Style.parse("blue"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_success_cleanup_event(*, dispatcher: Dispatcher,
                                                   director: DirectorPlugin,
                                                   reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        scenario_result1 = make_scenario_result().mark_passed() \
                                                 .set_started_at(1.0) \
                                                 .set_ended_at(2.0)
        scenario_result2 = make_scenario_result().mark_passed() \
                                                 .set_started_at(2.0) \
                                                 .set_ended_at(3.0)
        scenario_result3 = make_scenario_result().mark_skipped()
        report = make_report([
            scenario_result1,
            scenario_result2,
            scenario_result3,
        ])
        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(" "),
            call.out("# 3 scenarios, 2 passed, 0 failed, 1 skipped",
                     style=Style.parse("bold green"),
                     end=""),
            call.out(f" ({report.elapsed:.2f}s)", style=Style.parse("blue"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_failed_cleanup_event(*, dispatcher: Dispatcher,
                                                  director: DirectorPlugin,
                                                  reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        scenario_result = make_scenario_result().mark_failed() \
                                                .set_started_at(3.145) \
                                                .set_ended_at(6.285)
        report = make_report([scenario_result])
        event = CleanupEvent(report)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(" "),
            call.out("# 1 scenario, 0 passed, 1 failed, 0 skipped",
                     style=Style.parse("bold red"),
                     end=""),
            call.out(" (3.14s)", style=Style.parse("blue"))
        ]



from argparse import ArgumentParser, Namespace
from pathlib import Path
from unittest.mock import Mock

from baby_steps import given, then, when
from pytest import raises

from vedro.core import (
    Config,
    Event,
    ExcInfo,
    Report,
    ScenarioResult,
    StepResult,
    VirtualScenario,
    VirtualStep,
)
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    ExceptionRaisedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
    StepFailedEvent,
    StepPassedEvent,
    StepRunEvent,
)


def test_new_event():
    with when:
        class NewEvent(Event):
            pass

    with then:
        assert issubclass(NewEvent, Event)


def test_already_registered_event():
    with given:
        class RegisteredEvent(Event):
            pass

    with when, raises(BaseException) as exception:
        class RegisteredEvent(Event):  # noqa: F811
            pass

    with then:
        assert exception.type is Exception
        assert str(exception.value) == f"Event {RegisteredEvent!r} already registered"


def test_config_loaded_event():
    with given:
        path = Path()
        config = Config

    with when:
        event = ConfigLoadedEvent(path, config)

    with then:
        assert event.path == path
        assert event.config == config
        assert repr(event) == f"ConfigLoadedEvent({path!r}, <Config>)"


def test_arg_parse_event():
    with given:
        arg_parser = ArgumentParser()

    with when:
        event = ArgParseEvent(arg_parser)

    with then:
        assert event.arg_parser == arg_parser
        assert repr(event) == f"ArgParseEvent({arg_parser!r})"


def test_arg_parsed_event():
    with given:
        namespace = Namespace()

    with when:
        event = ArgParsedEvent(namespace)

    with then:
        assert event.args == namespace
        assert repr(event) == f"ArgParsedEvent({namespace!r})"


def test_startup_event():
    with given:
        scenarios = []

    with when:
        event = StartupEvent(scenarios)

    with then:
        assert event.scenarios == scenarios
        assert repr(event) == f"StartupEvent({scenarios!r})"


def test_scenario_run_event():
    with given:
        scenario_ = Mock(VirtualScenario)
        scenario_result = ScenarioResult(scenario_)

    with when:
        event = ScenarioRunEvent(scenario_result)

    with then:
        assert event.scenario_result == scenario_result
        assert repr(event) == f"ScenarioRunEvent({scenario_result!r})"


def test_scenario_skiped_event():
    with given:
        scenario_ = Mock(VirtualScenario)
        scenario_result = ScenarioResult(scenario_)

    with when:
        event = ScenarioSkippedEvent(scenario_result)

    with then:
        assert event.scenario_result == scenario_result
        assert repr(event) == f"ScenarioSkippedEvent({scenario_result!r})"


def test_scenario_failed_event():
    with given:
        scenario_ = Mock(VirtualScenario)
        scenario_result = ScenarioResult(scenario_)

    with when:
        event = ScenarioFailedEvent(scenario_result)

    with then:
        assert event.scenario_result == scenario_result
        assert repr(event) == f"ScenarioFailedEvent({scenario_result!r})"


def test_scenario_passed_event():
    with given:
        scenario_ = Mock(VirtualScenario)
        scenario_result = ScenarioResult(scenario_)

    with when:
        event = ScenarioPassedEvent(scenario_result)

    with then:
        assert event.scenario_result == scenario_result
        assert repr(event) == f"ScenarioPassedEvent({scenario_result!r})"


def test_step_run_event():
    with given:
        step_ = Mock(VirtualStep)
        step_result = StepResult(step_)

    with when:
        event = StepRunEvent(step_result)

    with then:
        assert event.step_result == step_result
        assert repr(event) == f"StepRunEvent({step_result!r})"


def test_step_fail_event():
    with given:
        step_ = Mock(VirtualStep)
        step_result = StepResult(step_)

    with when:
        event = StepFailedEvent(step_result)

    with then:
        assert event.step_result == step_result
        assert repr(event) == f"StepFailedEvent({step_result!r})"


def test_step_pass_event():
    with given:
        step_ = Mock(VirtualStep)
        step_result = StepResult(step_)

    with when:
        event = StepPassedEvent(step_result)

    with then:
        assert event.step_result == step_result
        assert repr(event) == f"StepPassedEvent({step_result!r})"


def test_exception_raised_event():
    with given:
        exception = AssertionError()
        exc_info = ExcInfo(type(exception), exception, None)

    with when:
        event = ExceptionRaisedEvent(exc_info)

    with then:
        assert event.exc_info == exc_info
        assert repr(event) == f"ExceptionRaisedEvent({exc_info!r})"


def test_cleanup_event():
    with given:
        report = Report()

    with when:
        event = CleanupEvent(report)

    with then:
        assert event.report == report
        assert repr(event) == f"CleanupEvent({report!r})"



from unittest.mock import sentinel

from baby_steps import given, then, when

from vedro.plugins.director import RichReporterPlugin
from vedro.plugins.director.rich.test_utils import console_, dispatcher, reporter

__all__ = ("dispatcher", "reporter", "console_",)


def test_rich_reporter_format_scope_without_values(*, reporter: RichReporterPlugin):
    with given:
        scope = {}

    with when:
        res = list(reporter._format_scope(scope))

    with then:
        assert res == []


def test_rich_reporter_format_scope_with_values(*, reporter: RichReporterPlugin):
    with given:
        scope = {"key_int": 1, "key_str": "val"}

    with when:
        res = list(reporter._format_scope(scope))

    with then:
        assert res == [
            ("key_int", "1"),
            ("key_str", '"val"'),
        ]


def test_rich_reporter_format_scope_with_unserializable_value(*, reporter: RichReporterPlugin):
    with given:
        unserializable = sentinel
        scope = {"key_int": 1, "key_unserializable": unserializable}

    with when:
        res = list(reporter._format_scope(scope))

    with then:
        assert res == [
            ("key_int", "1"),
            ("key_unserializable", repr(unserializable)),
        ]



import inspect
from functools import partialmethod
from typing import Any, Dict, Tuple


class _Meta(type):
    def __new__(mcs, name: str, bases: Tuple[Any], namespace: Dict[str, Any]) -> Any:
        if len(bases) == 0:
            return super().__new__(mcs, name, bases, namespace)

        for base in bases:
            if base != Scenario:
                module = namespace.get("__module__", "")
                raise TypeError(f"Subclassing is restricted <{module}.{name}>")

        cls_constructor = namespace.get("__init__")
        cls_params = getattr(cls_constructor, "__vedro__params__", None)
        if cls_params is None or len(cls_params) == 0:
            return super().__new__(mcs, name, bases, namespace)

        updated_name = "VedroTemplate"
        updated_namespace = {**namespace, "__qualname__": updated_name}
        created = super().__new__(mcs, updated_name, bases, updated_namespace)

        cls_globals = getattr(cls_constructor, "__globals__")
        for idx, (args, kwargs) in enumerate(reversed(cls_params), start=1):
            signature = inspect.signature(cls_constructor)  # type: ignore
            bound_args = signature.bind(None, *args, **kwargs)
            bound_args.apply_defaults()

            cls_name = f"{name}_{idx}_VedroScenario"
            cls_namespace = {
                **namespace,
                "__qualname__": cls_name,
                "__init__": partialmethod(cls_constructor, *args, **kwargs),
                "__vedro__template__": created,
                "__vedro__template_index__": idx,
                "__vedro__template_total__": len(cls_params),
                "__vedro__template_args__": bound_args,
            }
            cls_globals[cls_name] = type(cls_name, bases, cls_namespace)

        return created


class Scenario(metaclass=_Meta):
    subject: str

    def __repr__(self) -> str:
        return "<Scenario>"



import os
from hashlib import blake2b
from inspect import BoundArguments
from pathlib import Path
from typing import Any, List, Type, Union, cast

from .._scenario import Scenario
from ._virtual_step import VirtualStep

__all__ = ("VirtualScenario",)


class VirtualScenario:
    def __init__(self, orig_scenario: Type[Scenario], steps: List[VirtualStep]) -> None:
        self._orig_scenario = orig_scenario
        self._steps = steps
        self._path = Path(getattr(orig_scenario, "__file__", "."))
        self._is_skipped = False

    @property
    def steps(self) -> List[VirtualStep]:
        return self._steps

    @property
    def unique_id(self) -> str:
        unique_name = f"{self._path}::{self._orig_scenario.__qualname__}"
        return blake2b(unique_name.encode(), digest_size=20).hexdigest()

    @property
    def template_index(self) -> Union[int, None]:
        idx = getattr(self._orig_scenario, "__vedro__template_index__", None)
        return cast(Union[int, None], idx)

    @property
    def template_total(self) -> Union[int, None]:
        idx = getattr(self._orig_scenario, "__vedro__template_total__", None)
        return cast(Union[int, None], idx)

    @property
    def template_args(self) -> Union[BoundArguments, None]:
        args = getattr(self._orig_scenario, "__vedro__template_args__", None)
        return cast(Union[BoundArguments, None], args)

    @property
    def path(self) -> Path:
        return self._path

    @property
    def subject(self) -> str:
        subject = getattr(self._orig_scenario, "subject", None)
        if not isinstance(subject, str) or subject.strip() == "":
            subject = self._path.stem.replace("_", " ")

        if self.template_args:
            subject = subject.format(**self.template_args.arguments)
        return subject

    @property
    def namespace(self) -> str:
        module = self._orig_scenario.__module__
        rel_path = os.path.relpath(self._path, module.split(".")[0])
        return os.path.dirname(rel_path)

    def skip(self) -> None:
        self._is_skipped = True

    def is_skipped(self) -> bool:
        return self._is_skipped

    def __call__(self) -> Scenario:
        return self._orig_scenario()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._orig_scenario!r}, {self._steps!r})"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



from pytest import raises

from vedro.core._scenario_finder._file_filters import FileFilter


def test_file_filter():
    with raises(Exception) as exc_info:
        FileFilter()

    assert exc_info.type is TypeError
    assert "Can't instantiate abstract class FileFilter" in str(exc_info.value)



from pathlib import Path

from baby_steps import given, then, when
from pytest import raises

from vedro.core import FileArtifact, MemoryArtifact


def test_memory_artifact():
    with given:
        name = "log"
        mime_type = "text/plain"
        data = b""

    with when:
        artifact = MemoryArtifact(name, mime_type, data)

    with then:
        assert artifact.name == name
        assert artifact.mime_type == mime_type
        assert artifact.data == data


def test_memory_artifact_repr():
    with given:
        name = "log"
        mime_type = "text/plain"
        data = b""

    with when:
        artifact = MemoryArtifact(name, mime_type, data)

    with then:
        assert repr(artifact) == f"MemoryArtifact<{name!r}, {mime_type!r}, size={len(data)}>"


def test_memory_artifact_binary_only():
    with raises(Exception) as exc_info:
        MemoryArtifact("log", "text/plain", "text")

    with then:
        assert exc_info.type is AssertionError


def test_file_artifact():
    with given:
        name = "log"
        mime_type = "text/plain"
        path = Path()

    with when:
        artifact = FileArtifact(name, mime_type, path)

    with then:
        assert artifact.name == name
        assert artifact.mime_type == mime_type
        assert artifact.path == path


def test_file_artifact_repr():
    with given:
        name = "log"
        mime_type = "text/plain"
        path = Path()

    with when:
        artifact = FileArtifact(name, mime_type, path)

    with then:
        print(repr(artifact))
        assert repr(artifact) == f"FileArtifact<{name!r}, {mime_type!r}, {path!r}>"


def test_file_artifact_path_only():
    with raises(Exception) as exc_info:
        FileArtifact("log", "text/plain", "./log.txt")

    with then:
        assert exc_info.type is AssertionError



import os
from argparse import HelpFormatter
from functools import partial
from pathlib import Path
from typing import Tuple

from ..events import ArgParsedEvent, ArgParseEvent, CleanupEvent, ConfigLoadedEvent, StartupEvent
from ._arg_parser import ArgumentParser
from ._config_loader import ConfigLoader, ConfigType
from ._dispatcher import Dispatcher
from ._plugin import Plugin
from ._report import Report
from ._runner import Runner
from ._scenario_discoverer import ScenarioDiscoverer

__all__ = ("Lifecycle",)


class Lifecycle:
    def __init__(self,
                 dispatcher: Dispatcher,
                 discoverer: ScenarioDiscoverer,
                 runner: Runner,
                 config_loader: ConfigLoader) -> None:
        self._dispatcher = dispatcher
        self._discoverer = discoverer
        self._runner = runner
        self._config_loader = config_loader

    async def _load_config(self, filename: Path) -> Tuple[Path, ConfigType]:
        parser = ArgumentParser(add_help=False)
        parser.add_argument("--config", default=filename, type=Path)

        args, _ = parser.parse_known_args()
        if args.config != filename:
            assert args.config.exists(), f"'{args.config}' does not exist"

        config = await self._config_loader.load(args.config)
        return args.config, config

    async def start(self) -> Report:
        formatter = partial(HelpFormatter, max_help_position=30)
        default_config = Path("vedro.cfg.py")

        arg_parser = ArgumentParser("vedro", formatter_class=formatter, add_help=False,
                                    description="documentation: vedro.io/docs")
        arg_parser.add_argument("--config", default=default_config, type=Path,
                                help=f"Config path (default: {default_config})")
        arg_parser.add_argument("-h", "--help",
                                action="help", help="Show this help message and exit")

        config_path, config = await self._load_config(default_config)
        for _, section in config.Plugins.items():
            assert issubclass(section.plugin, Plugin)
            assert section.plugin != Plugin
            if section.enabled:
                plugin = section.plugin(config=section)
                self._dispatcher.register(plugin)
        await self._dispatcher.fire(ConfigLoadedEvent(config_path, config))

        subparsers = arg_parser.add_subparsers(dest="subparser")
        arg_parser_run = subparsers.add_parser("run", add_help=False,
                                               description="documentation: vedro.io/docs",
                                               help="Run scenarios. "
                                                    "Type 'vedro run --help' for more info")
        arg_parser.set_default_subparser("run")

        await self._dispatcher.fire(ArgParseEvent(arg_parser_run))
        arg_parser_run.add_argument("--reruns", type=int, default=0,
                                    help="Number of times to rerun failed scenarios (default: 0)")
        arg_parser_run.add_argument("--config", default=default_config, type=Path,
                                    help=f"Config path (default: {default_config})")
        arg_parser_run.add_argument("-h", "--help",
                                    action="help", help="Show this help message and exit")

        args = arg_parser.parse_args()
        await self._dispatcher.fire(ArgParsedEvent(args))

        start_dir = os.path.relpath(Path("scenarios"))
        scenarios = await self._discoverer.discover(Path(start_dir))
        await self._dispatcher.fire(StartupEvent(scenarios))

        report = await self._runner.run(scenarios, reruns=args.reruns)

        await self._dispatcher.fire(CleanupEvent(report))

        return report

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return (f"{cls_name}({self._dispatcher!r}, {self._discoverer!r}, "
                f"{self._runner!r}, {self._config_loader!r})")



from pytest import raises

from vedro.core import ScenarioFinder


def test_scenario_finder():
    with raises(Exception) as exc_info:
        ScenarioFinder()

    assert exc_info.type is TypeError
    assert "Can't instantiate abstract class ScenarioFinder" in str(exc_info.value)



import json
import os
from traceback import format_exception
from types import TracebackType
from typing import Any, Callable, Dict, Generator, Tuple, Type

from rich.console import Console
from rich.style import Style

import vedro
from vedro.core import Dispatcher, PluginConfig, ScenarioResult
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)

from .._director_init_event import DirectorInitEvent
from .._reporter import Reporter
from ..rich.utils import make_console

__all__ = ("PyCharmReporter", "PyCharmReporterPlugin",)


class PyCharmReporterPlugin(Reporter):
    def __init__(self, config: Type["PyCharmReporter"], *,
                 console_factory: Callable[[], Console] = make_console) -> None:
        super().__init__(config)
        self._console = console_factory()
        self._show_internal_calls = config.show_internal_calls
        self._show_skipped = config.show_skipped

    def subscribe(self, dispatcher: Dispatcher) -> None:
        super().subscribe(dispatcher)
        dispatcher.listen(DirectorInitEvent, lambda e: e.director.register("pycharm", self))

    def on_chosen(self) -> None:
        assert isinstance(self._dispatcher, Dispatcher)
        self._dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                        .listen(ArgParsedEvent, self.on_arg_parsed) \
                        .listen(StartupEvent, self.on_startup) \
                        .listen(ScenarioRunEvent, self.on_scenario_run) \
                        .listen(ScenarioPassedEvent, self.on_scenario_passed) \
                        .listen(ScenarioFailedEvent, self.on_scenario_failed) \
                        .listen(ScenarioSkippedEvent, self.on_scenario_skipped) \
                        .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("PyCharm Reporter")

        group.add_argument("--pycharm-show-skipped",
                           action="store_true",
                           default=self._show_skipped,
                           help="Show skipped scenarios")
        group.add_argument("--pycharm-show-internal-calls",
                           action="store_true",
                           default=self._show_internal_calls,
                           help="Show internal calls in the traceback output")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._show_internal_calls = event.args.pycharm_show_internal_calls
        self._show_skipped = event.args.pycharm_show_skipped

    def on_startup(self, event: StartupEvent) -> None:
        self._console.out("Scenarios")

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        scenario_result: ScenarioResult = event.scenario_result
        self._write_message("testStarted", {
            "name": scenario_result.scenario.subject,
            "locationHint": scenario_result.scenario.path,
        })

    def on_scenario_passed(self, event: ScenarioPassedEvent) -> None:
        scenario_result: ScenarioResult = event.scenario_result

        subject = f"✔ {scenario_result.scenario.subject}"
        self._console.out(subject, style=Style(color="green"))

        self._write_message("testFinished", {
            "name": scenario_result.scenario.subject,
            "duration": int(scenario_result.elapsed * 1000),
        })

    def on_scenario_failed(self, event: ScenarioFailedEvent) -> None:
        scenario_result: ScenarioResult = event.scenario_result

        subject = f"✗ {scenario_result.scenario.subject}"
        self._console.out(subject, style=Style(color="red"))

        for step_result in scenario_result.step_results:
            if step_result.exc_info:
                self._print_exception(step_result.exc_info.value, step_result.exc_info.traceback)

        if scenario_result.scope:
            self._print_scope(scenario_result.scope)

        self._write_message("testFailed", {
            "name": scenario_result.scenario.subject,
            "message": "",
            "details": "",
        })
        self._write_message("testFinished", {
            "name": scenario_result.scenario.subject,
            "duration": int(scenario_result.elapsed * 1000),
        })

    def on_scenario_skipped(self, event: ScenarioSkippedEvent) -> None:
        scenario_result: ScenarioResult = event.scenario_result
        if self._show_skipped:
            self._write_message("testIgnored", {
                "name": scenario_result.scenario.subject,
            })

    def on_cleanup(self, event: CleanupEvent) -> None:
        pass

    def _print_exception(self, exception: BaseException, traceback: TracebackType) -> None:
        if not self._show_internal_calls:
            root = os.path.dirname(vedro.__file__)
            while traceback.tb_next is not None:
                filename = os.path.abspath(traceback.tb_frame.f_code.co_filename)
                if os.path.commonpath([root, filename]) != root:
                    break
                traceback = traceback.tb_next

        formatted = format_exception(type(exception), exception, traceback)
        self._console.out("".join(formatted), style=Style(color="yellow"))

    def _print_scope(self, scope: Dict[Any, Any]) -> None:
        self._console.out("Scope:", style=Style(color="blue", bold=True))
        for key, val in self._format_scope(scope):
            self._console.out(f" {key}: ", end="", style=Style(color="blue"))
            self._console.out(val)

    def _format_scope(self, scope: Dict[Any, Any]) -> Generator[Tuple[str, str], None, None]:
        for key, val in scope.items():
            try:
                val_repr = json.dumps(val, ensure_ascii=False, indent=4)
            except:  # noqa: E722
                val_repr = repr(val)
            yield str(key), val_repr

    def _escape_value(self, value: str) -> str:
        symbols = {"'": "|'", "\n": "|n", "\r": "|r", "|": "||", "[": "|[", "]": "|]"}
        return value.translate({ord(k): v for k, v in symbols.items()})

    def _write_message(self, name: str, attributes: Dict[str, Any]) -> None:
        message = f"##teamcity[{name}"
        for key, val in attributes.items():
            escaped = self._escape_value(str(val))
            message += f" {key}='{escaped}'"
        message += "]"
        self._console.out(message)


class PyCharmReporter(PluginConfig):
    plugin = PyCharmReporterPlugin

    # Show internal calls in the traceback output
    show_internal_calls: bool = False

    # Show skipped scenarios
    show_skipped: bool = False



import importlib
import importlib.util
from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from pathlib import Path
from types import ModuleType
from typing import cast

from ._module_loader import ModuleLoader

__all__ = ("ModuleFileLoader",)


class ModuleFileLoader(ModuleLoader):
    def _path_to_module_name(self, path: Path) -> str:
        return ".".join(path.with_suffix("").parts)

    def _spec_from_path(self, path: Path) -> ModuleSpec:
        module_name = self._path_to_module_name(path)
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None:
            raise ModuleNotFoundError(module_name)
        return spec

    def _module_from_spec(self, spec: ModuleSpec) -> ModuleType:
        return importlib.util.module_from_spec(spec)

    def _exec_module(self, loader: Loader, module: ModuleType) -> None:
        loader.exec_module(module)

    async def load(self, path: Path) -> ModuleType:
        spec = self._spec_from_path(path)
        module = self._module_from_spec(spec)
        self._exec_module(cast(Loader, spec.loader), module)
        return module



import os
import random
import string
import sys
from argparse import Namespace
from pathlib import Path
from types import TracebackType
from typing import Any, List, Optional, cast
from unittest.mock import Mock

import pytest
from rich.console import Console

import vedro.plugins.director as d
from vedro import Scenario
from vedro.core import (
    ArgumentParser,
    Config,
    Dispatcher,
    ExcInfo,
    Report,
    ScenarioResult,
    StepResult,
    VirtualScenario,
    VirtualStep,
)
from vedro.events import ArgParseEvent, ConfigLoadedEvent
from vedro.plugins.director import Director, DirectorPlugin, RichReporterPlugin

__all__ = ("dispatcher", "console_", "reporter", "director", "chose_reporter",
           "make_parsed_args", "make_path", "make_vscenario", "make_vstep",
           "make_scenario_result", "make_step_result", "make_random_name", "make_exc_info",
           "make_report",)


@pytest.fixture()
def dispatcher() -> Dispatcher:
    return Dispatcher()


@pytest.fixture()
def director(dispatcher: Dispatcher) -> DirectorPlugin:
    director = DirectorPlugin(Director)
    director.subscribe(dispatcher)
    return director


@pytest.fixture()
def console_() -> Console:
    return Mock(Console)


@pytest.fixture()
def reporter(dispatcher: Dispatcher, console_: Console) -> RichReporterPlugin:
    class RichReporter(d.RichReporter):
        tb_pretty = False
    reporter = RichReporterPlugin(RichReporter, console_factory=lambda: console_)
    reporter.subscribe(dispatcher)
    return reporter


async def chose_reporter(dispatcher: Dispatcher,
                         director: DirectorPlugin, reporter: RichReporterPlugin) -> None:
    await dispatcher.fire(ConfigLoadedEvent(Path(), Config))
    await dispatcher.fire(ArgParseEvent(ArgumentParser()))


def make_parsed_args(*, verbose: int = 0,
                     show_timings: bool = False,
                     show_paths: bool = False,
                     tb_show_locals: bool = False,
                     tb_show_internal_calls: bool = False,
                     reruns: int = 0) -> Namespace:
    return Namespace(
        verbose=verbose,
        show_timings=show_timings,
        show_paths=show_paths,
        tb_show_internal_calls=tb_show_internal_calls,
        tb_show_locals=tb_show_locals,
        reruns=reruns,
    )


def make_path(path: str = "", name: str = "scenario.py") -> Path:
    return Path(os.getcwd()) / "scenarios" / path / name


def make_vscenario(*,
                   path: Optional[Path] = None,
                   subject: Optional[str] = None) -> VirtualScenario:
    namespace = {}
    if path:
        namespace["__file__"] = str(path)
        rel_path = path.relative_to(os.getcwd())
        namespace["__module__"] = str(rel_path.with_suffix("")).replace("/", ".")
    if subject:
        namespace["subject"] = subject
    scenario = type("Scenario", (Scenario,), namespace)

    vscenario = VirtualScenario(scenario, [])
    return vscenario


def make_vstep(*, name: Optional[str] = None) -> VirtualStep:
    def method(self: Any) -> None:
        pass
    if name:
        method.__name__ = name
    return VirtualStep(method)


def make_random_name(length: int = 10) -> str:
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(length))


def make_scenario_result(vscenario: Optional[VirtualScenario] = None,
                         step_results: Optional[List[StepResult]] = None) -> ScenarioResult:
    if vscenario is None:
        vscenario = make_vscenario(subject=make_random_name())
    scenario_result = ScenarioResult(vscenario)

    if step_results:
        for step_result in step_results:
            scenario_result.add_step_result(step_result)
    return scenario_result


def make_step_result(vstep: Optional[VirtualStep] = None) -> StepResult:
    if vstep is None:
        vstep = make_vstep(name=make_random_name())
    step_result = StepResult(vstep)
    return step_result


def make_exc_info(exc_val: Exception) -> ExcInfo:
    try:
        raise exc_val
    except type(exc_val):
        *_, traceback = sys.exc_info()
    return ExcInfo(type(exc_val), exc_val, cast(TracebackType, traceback))


def make_report(scenario_results: Optional[List[ScenarioResult]] = None,
                summaries: Optional[List[str]] = None) -> Report:
    report = Report()
    if scenario_results:
        for scenarior_result in scenario_results:
            report.add_result(scenarior_result)
    if summaries:
        for summary in summaries:
            report.add_summary(summary)
    return report



from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when

from vedro.core import Dispatcher
from vedro.events import StartupEvent
from vedro.plugins.director import DirectorPlugin, RichReporterPlugin
from vedro.plugins.director.rich.test_utils import (
    chose_reporter,
    console_,
    director,
    dispatcher,
    reporter,
)

__all__ = ("dispatcher", "reporter", "director", "console_",)


@pytest.mark.asyncio
async def test_rich_reporter_startup_event(*, dispatcher: Dispatcher, director: DirectorPlugin,
                                           reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)
        event = StartupEvent([])

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out("Scenarios")
        ]



from typing import Type

from ._config_loader import Section
from ._dispatcher import Dispatcher, Subscriber

__all__ = ("Plugin", "PluginConfig",)


class Plugin(Subscriber):
    def __init__(self, config: Type["PluginConfig"]) -> None:
        self._config = config

    def subscribe(self, dispatcher: Dispatcher) -> None:
        pass


class PluginConfig(Section):
    plugin: Type[Plugin] = Plugin
    enabled: bool = True



from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when
from rich.console import Style

from vedro.core import Dispatcher, ScenarioResult
from vedro.events import ScenarioRunEvent
from vedro.plugins.director import DirectorPlugin, RichReporterPlugin
from vedro.plugins.director.rich.test_utils import (
    chose_reporter,
    console_,
    director,
    dispatcher,
    make_path,
    make_scenario_result,
    make_vscenario,
    reporter,
)

__all__ = ("dispatcher", "reporter", "director", "console_",)


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event(*, dispatcher: Dispatcher,
                                                director: DirectorPlugin,
                                                reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        vscenario = make_vscenario(path=make_path("<namespace_1>"))
        scenario_result = ScenarioResult(vscenario)
        event = ScenarioRunEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out("* <namespace 1>", style=Style.parse("bold"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event_same_namespace(*, dispatcher: Dispatcher,
                                                               director: DirectorPlugin,
                                                               reporter: RichReporterPlugin,
                                                               console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        vscenario = make_vscenario(path=make_path("<namespace_1>"))
        scenario_result = make_scenario_result(vscenario)
        event = ScenarioRunEvent(scenario_result)

        await dispatcher.fire(event)
        console_.reset_mock()

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_scenario_run_event_diff_namespace(*, dispatcher: Dispatcher,
                                                               director: DirectorPlugin,
                                                               reporter: RichReporterPlugin,
                                                               console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        vscenario = make_vscenario(path=make_path("<namespace_1>"))
        scenario_result1 = ScenarioResult(vscenario)
        event1 = ScenarioRunEvent(scenario_result1)
        await dispatcher.fire(event1)
        console_.reset_mock()

        vscenario = make_vscenario(path=make_path("<namespace_2>"))
        scenario_result2 = ScenarioResult(vscenario)
        event2 = ScenarioRunEvent(scenario_result2)

    with when:
        await dispatcher.fire(event2)

    with then:
        assert console_.mock_calls == [
            call.out("* <namespace 2>", style=Style.parse("bold")),
        ]



from argparse import ArgumentParser
from pathlib import Path
from unittest.mock import Mock

import pytest
from baby_steps import given, then, when

from vedro import Config
from vedro.core import Dispatcher
from vedro.events import ArgParseEvent, ConfigLoadedEvent
from vedro.plugins.director import DirectorPlugin, RichReporterPlugin
from vedro.plugins.director.rich.test_utils import console_, director, dispatcher, reporter

__all__ = ("dispatcher", "reporter", "director", "console_",)


@pytest.mark.asyncio
async def test_rich_reporter_arg_parse_event(*, dispatcher: Dispatcher, director: DirectorPlugin,
                                             reporter: RichReporterPlugin, console_: Mock):
    with given:
        await dispatcher.fire(ConfigLoadedEvent(Path(), Config))

        parser = ArgumentParser()
        event = ArgParseEvent(parser)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []



from typing import Any, Callable

__all__ = ("params",)


class params:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._args = args
        self._kwargs = kwargs

    def __call__(self, fn: Callable[..., Any]) -> Callable[..., Any]:
        if not hasattr(fn, "__vedro__params__"):
            setattr(fn, "__vedro__params__", [])
        getattr(fn, "__vedro__params__").append((self._args, self._kwargs))
        return fn



from pathlib import Path
from typing import Optional

from .._module_loader import ModuleFileLoader, ModuleLoader
from ._config_loader import ConfigLoader
from ._config_type import Config, ConfigType

__all__ = ("ConfigFileLoader",)


class ConfigFileLoader(ConfigLoader):
    def __init__(self, default_config: ConfigType,
                 module_loader: Optional[ModuleLoader] = None) -> None:
        super().__init__(default_config)
        self._module_loader = module_loader or ModuleFileLoader()

    async def load(self, path: Path) -> ConfigType:
        if not path.exists():
            return self._default_config
        module = await self._module_loader.load(path)

        config = getattr(module, "Config", self._default_config)
        assert issubclass(config, Config)

        return config



from pathlib import Path
from textwrap import dedent

import pytest
from baby_steps import given, then, when

from vedro.core._scenario_loader import ScenarioFileLoader


@pytest.mark.asyncio
@pytest.mark.parametrize("name", [
    "Scenario",  # default
    "TmpScenario",  # custom name
    "Scenario_1",  # backward compatibility
    "Scenario_1_VedroScenario",  # parametrized
])
async def test_scenario_file_loader(name, *, tmp_path: Path):
    with given:
        path = tmp_path / "scenario.py"
        path.write_text(dedent(f'''
            import vedro
            class {name}(vedro.Scenario):
                pass
        '''))

        loader = ScenarioFileLoader()

    with when:
        scenarios = await loader.load(path)

    with then:
        assert len(scenarios) == 1


@pytest.mark.asyncio
async def test_scenarios_file_loader(tmp_path: Path):
    with given:
        path = tmp_path / "scenario.py"
        path.write_text(dedent('''
            import vedro
            class CreateUserScenario(vedro.Scenario):
                pass
            class UpdatedUserScenario(vedro.Scenario):
                pass
        '''))

        loader = ScenarioFileLoader()

    with when:
        scenarios = await loader.load(path)

    with then:
        assert len(scenarios) == 2


@pytest.mark.asyncio
async def test_template_scenario_file_loader(tmp_path: Path):
    with given:
        path = tmp_path / "scenario.py"
        path.write_text(dedent('''
            import vedro
            class Scenario(vedro.Scenario):
                @vedro.params("Bob")
                @vedro.params("Alice")
                def __init__(self, user):
                    pass
        '''))

        loader = ScenarioFileLoader()

    with when:
        scenarios = await loader.load(path)

    with then:
        assert len(scenarios) == 2



import os
from pathlib import Path
from types import MethodType
from typing import Type
from unittest.mock import Mock

import pytest
from baby_steps import given, then, when

from vedro import Scenario
from vedro.core import (
    MemoryArtifact,
    ScenarioResult,
    ScenarioStatus,
    StepResult,
    VirtualScenario,
    VirtualStep,
)


def make_scenario_path(path: str = "", name: str = "scenario.py") -> Path:
    return Path(os.getcwd()) / "scenarios" / path / name


@pytest.fixture()
def scenario_():
    scenario = Mock(Scenario)
    scenario.__file__ = str(make_scenario_path())
    return scenario


@pytest.fixture()
def virtual_scenario(scenario_: Type[scenario_]):
    virtual_scenario = VirtualScenario(scenario_, [])
    return virtual_scenario


def test_scenario_result():
    with when:
        subject = "<subject>"
        namespace = "<namespace>"

        scenario_ = Mock(Scenario)
        scenario_.subject = subject
        scenario_.__file__ = str(make_scenario_path(namespace))
        virtual_scenario = VirtualScenario(scenario_, [])

        scenario_result = ScenarioResult(virtual_scenario)

    with then:
        assert scenario_result.scenario == virtual_scenario
        assert scenario_result.step_results == []
        assert scenario_result.is_passed() is False
        assert scenario_result.is_failed() is False
        assert scenario_result.is_skipped() is False
        assert scenario_result.started_at is None
        assert scenario_result.ended_at is None
        assert scenario_result.scope == {}
        assert scenario_result.rerun == 0
        assert scenario_result.status == ScenarioStatus.PENDING


def test_scenario_result_mark_passed(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)

    with when:
        res = scenario_result.mark_passed()

    with then:
        assert res == scenario_result
        assert scenario_result.is_passed() is True


def test_scenario_result_mark_failed(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)

    with when:
        res = scenario_result.mark_failed()

    with then:
        assert res == scenario_result
        assert scenario_result.is_failed() is True


def test_scenario_result_mark_skipped(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)

    with when:
        res = scenario_result.mark_skipped()

    with then:
        assert res == scenario_result
        assert scenario_result.is_skipped() is True


def test_scenario_result_set_started_at(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)
        started_at = 1.0

    with when:
        res = scenario_result.set_started_at(started_at)

    with then:
        assert res == scenario_result
        assert scenario_result.started_at == started_at
        assert scenario_result.elapsed == 0.0


def test_scenario_result_set_ended_at(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)
        ended_at = 1.0

    with when:
        res = scenario_result.set_ended_at(ended_at)

    with then:
        assert res == scenario_result
        assert scenario_result.ended_at == ended_at
        assert scenario_result.elapsed == 0.0


def test_scenario_result_elapsed(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)
        started_at = 3.0
        scenario_result.set_started_at(started_at)
        ended_at = 1.0
        scenario_result.set_ended_at(ended_at)

    with when:
        res = scenario_result.elapsed

    with then:
        assert res == ended_at - started_at


def test_scenario_result_add_step_result(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)
        virtual_step = VirtualStep(Mock(MethodType))
        step_result = StepResult(virtual_step)

    with when:
        res = scenario_result.add_step_result(step_result)

    with then:
        assert res is None
        assert scenario_result.step_results == [step_result]


def test_scenario_result_get_step_results(*, virtual_scenario: VirtualScenario):
    with given:
        virtual_step1 = VirtualStep(Mock(MethodType))
        step_result1 = StepResult(virtual_step1)

        virtual_step2 = VirtualStep(Mock(MethodType))
        step_result2 = StepResult(virtual_step2)

        scenario_result = ScenarioResult(virtual_scenario)
        scenario_result.add_step_result(step_result1)
        scenario_result.add_step_result(step_result2)

    with when:
        step_results = scenario_result.step_results

    with then:
        assert step_results == [step_result1, step_result2]


def test_scenario_result_set_scope(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)
        scope = {}

    with when:
        res = scenario_result.set_scope(scope)

    with then:
        assert res is None
        assert scenario_result.scope == scope


def test_scenario_result_repr(*, virtual_scenario: VirtualScenario):
    with when:
        scenario_result = ScenarioResult(virtual_scenario)

    with then:
        assert repr(scenario_result) == f"ScenarioResult<{virtual_scenario!r}>"


def test_scenario_result_eq(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result1 = ScenarioResult(virtual_scenario)
        scenario_result2 = ScenarioResult(virtual_scenario)

    with when:
        res = scenario_result1 == scenario_result2

    with then:
        assert res is True


def test_scenario_result_not_eq():
    with given:
        scenario1_ = Mock(Scenario)
        scenario1_.__file__ = str(make_scenario_path())
        virtual_scenario1 = VirtualScenario(scenario1_, [])
        scenario_result1 = ScenarioResult(virtual_scenario1)

        scenario2_ = Mock(Scenario)
        scenario2_.__file__ = str(make_scenario_path())
        virtual_scenario2 = VirtualScenario(scenario2_, [])
        scenario_result2 = ScenarioResult(virtual_scenario2)

    with when:
        res = scenario_result1 == scenario_result2

    with then:
        assert res is False


def test_scenario_result_attach_artifact(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)
        artifact = MemoryArtifact("name", "text/plain", b"")

    with when:
        res = scenario_result.attach(artifact)

    with then:
        assert res is None


def test_scenario_result_get_artifacts(*, virtual_scenario: VirtualScenario):
    with given:
        scenario_result = ScenarioResult(virtual_scenario)

        artifact1 = MemoryArtifact("name1", "text/plain", b"")
        scenario_result.attach(artifact1)

        artifact2 = MemoryArtifact("name2", "text/plain", b"")
        scenario_result.attach(artifact2)

    with when:
        artifacts = scenario_result.artifacts

    with then:
        assert artifacts == [artifact1, artifact2]



from pathlib import Path
from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when
from rich.console import Style

from vedro.core import Dispatcher, ScenarioResult
from vedro.events import ArgParsedEvent, ScenarioPassedEvent
from vedro.plugins.director import DirectorPlugin, RichReporterPlugin
from vedro.plugins.director.rich.test_utils import (
    chose_reporter,
    console_,
    director,
    dispatcher,
    make_parsed_args,
    make_vscenario,
    reporter,
)

__all__ = ("dispatcher", "reporter", "director", "console_",)


@pytest.mark.asyncio
async def test_rich_reporter_scenario_passed_event(*, dispatcher: Dispatcher,
                                                   director: DirectorPlugin,
                                                   reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        subject = "<subject>"
        vscenario = make_vscenario(subject=subject)
        scenario_result = ScenarioResult(vscenario).mark_passed()
        event = ScenarioPassedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✔ {subject}", style=Style.parse("green"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_passed_event_with_show_paths(*,
                                                                   dispatcher: Dispatcher,
                                                                   director: DirectorPlugin,
                                                                   reporter: RichReporterPlugin,
                                                                   console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        args = make_parsed_args(show_paths=True)
        await dispatcher.fire(ArgParsedEvent(args))

        subject = "<subject>"
        path = "scenarios/scenario.py"
        vscenario = make_vscenario(subject=subject, path=Path(path).absolute())
        scenario_result = ScenarioResult(vscenario).mark_passed()
        event = ScenarioPassedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✔ {subject}", style=Style.parse("green")),
            call.out(f"   > {path}", style=Style.parse("grey50")),
        ]



import sys
from typing import Any, List, Type

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from unittest.mock import Mock

import pytest
from baby_steps import given, then, when

from vedro import Scenario
from vedro.core import Dispatcher, Runner, VirtualScenario, VirtualStep
from vedro.events import (
    ExceptionRaisedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    StepFailedEvent,
    StepPassedEvent,
    StepRunEvent,
)


@pytest.fixture()
def dispatcher_():
    return AsyncMock(Dispatcher())


@pytest.fixture()
def runner(dispatcher_: Dispatcher):
    return Runner(dispatcher_)


def assert_mock_calls(mock_calls: List[Any], expected: List[Any]) -> bool:
    assert len(mock_calls) == len(expected)

    for idx, exp in enumerate(expected):
        name, args, kwargs = mock_calls[idx]
        assert len(args) == 1
        assert isinstance(args[0], exp)

    return True


def make_scenario(steps: List[Any] = None) -> Type[Scenario]:
    kwargs = {}
    if steps is not None:
        for idx, step in enumerate(steps):
            kwargs[f"step_{idx}"] = step
    return Mock(Scenario, **kwargs, __file__="/tmp/scenario.py")


@pytest.mark.asyncio
@pytest.mark.parametrize("reruns", (0, 1))
async def test_runner_run_scenario(reruns: int, *, runner: Runner, dispatcher_: Dispatcher):
    with given:
        virtual_scenario = VirtualScenario(make_scenario(), [])

    with when:
        report = await runner.run([virtual_scenario], reruns=reruns)

    with then:
        assert report.total == report.passed == 1
        assert assert_mock_calls(dispatcher_.mock_calls, [
            ScenarioRunEvent,
            ScenarioPassedEvent,
        ])


@pytest.mark.asyncio
@pytest.mark.parametrize("reruns", (0, 1, 3))
async def test_runner_run_scenario_all_failed(reruns: int, *,
                                              runner: Runner, dispatcher_: Dispatcher):
    with given:
        step_ = Mock(side_effect=AssertionError())
        scenario_ = make_scenario(steps=[step_])

        step = VirtualStep(step_)
        virtual_scenario = VirtualScenario(scenario_, [step])

    with when:
        report = await runner.run([virtual_scenario], reruns=reruns)

    with then:
        assert report.total == report.failed == 1
        assert assert_mock_calls(dispatcher_.mock_calls, [
            ScenarioRunEvent,
            StepRunEvent,
            ExceptionRaisedEvent,
            StepFailedEvent,
            ScenarioFailedEvent,
        ] * (reruns + 1))


@pytest.mark.asyncio
async def test_runner_run_scenario_most_failed(runner: Runner, dispatcher_: Dispatcher):
    with given:
        step_ = Mock(side_effect=(AssertionError(), None, AssertionError()))
        scenario_ = make_scenario(steps=[step_])

        step = VirtualStep(step_)
        virtual_scenario = VirtualScenario(scenario_, [step])

    with when:
        report = await runner.run([virtual_scenario], reruns=2)

    with then:
        assert report.total == report.failed == 1
        assert assert_mock_calls(dispatcher_.mock_calls, [
            ScenarioRunEvent,
            StepRunEvent,
            ExceptionRaisedEvent,
            StepFailedEvent,
            ScenarioFailedEvent,

            ScenarioRunEvent,
            StepRunEvent,
            StepPassedEvent,
            ScenarioPassedEvent,

            ScenarioRunEvent,
            StepRunEvent,
            ExceptionRaisedEvent,
            StepFailedEvent,
            ScenarioFailedEvent,
        ])


@pytest.mark.asyncio
async def test_runner_run_scenario_most_passed(runner: Runner, dispatcher_: Dispatcher):
    with given:
        step_ = Mock(side_effect=(AssertionError(), None, None))
        scenario_ = make_scenario(steps=[step_])

        step = VirtualStep(step_)
        virtual_scenario = VirtualScenario(scenario_, [step])

    with when:
        report = await runner.run([virtual_scenario], reruns=2)

    with then:
        assert report.total == report.passed == 1
        assert assert_mock_calls(dispatcher_.mock_calls, [
            ScenarioRunEvent,
            StepRunEvent,
            ExceptionRaisedEvent,
            StepFailedEvent,
            ScenarioFailedEvent,

            ScenarioRunEvent,
            StepRunEvent,
            StepPassedEvent,
            ScenarioPassedEvent,

            ScenarioRunEvent,
            StepRunEvent,
            StepPassedEvent,
            ScenarioPassedEvent,
        ])



import importlib
import importlib.util
import inspect
import os
from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from pathlib import Path
from types import ModuleType
from typing import List, Type, cast

from ..._scenario import Scenario
from ._scenario_loader import ScenarioLoader

__all__ = ("ScenarioFileLoader",)


class ScenarioFileLoader(ScenarioLoader):
    def _path_to_module_name(self, path: Path) -> str:
        return ".".join(path.with_suffix("").parts)

    def _spec_from_path(self, path: Path) -> ModuleSpec:
        module_name = self._path_to_module_name(path)
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None:
            raise ModuleNotFoundError(module_name)
        return spec

    def _module_from_spec(self, spec: ModuleSpec) -> ModuleType:
        return importlib.util.module_from_spec(spec)

    def _exec_module(self, loader: Loader, module: ModuleType) -> None:
        loader.exec_module(module)

    async def load(self, path: Path) -> List[Type[Scenario]]:
        spec = self._spec_from_path(path)
        module = self._module_from_spec(spec)
        self._exec_module(cast(Loader, spec.loader), module)

        loaded = []
        for name in dir(module):
            val = getattr(module, name)
            if inspect.isclass(val) and issubclass(val, Scenario) and val != Scenario:
                if not (val.__name__.startswith("Scenario") or val.__name__.endswith("Scenario")):
                    continue
                val.__file__ = os.path.abspath(module.__file__)  # type: ignore
                loaded.append(val)
        return loaded



import os
import shutil
from typing import Any

from rich.console import Console

__all__ = ("make_console", "get_terminal_size",)


def get_terminal_size(default_columns: int = 80, default_lines: int = 24) -> os.terminal_size:
    columns, lines = shutil.get_terminal_size()
    # Fix columns=0 lines=0 in Pycharm
    if columns <= 0:
        columns = default_columns
    if lines <= 0:
        lines = default_lines
    return os.terminal_size((columns, lines))


def make_console(**options: Any) -> Console:
    size = get_terminal_size()
    default_options = dict(
        highlight=False,
        force_terminal=True,
        width=size.columns,
        height=size.lines,
    )
    return Console(**{**default_options, **options})  # type: ignore



import asyncio
import os
import sys
from asyncio import CancelledError
from typing import List, Optional

from ._config import Config
from ._context import context
from ._interface import Interface
from ._params import params
from ._scenario import Scenario
from ._version import version
from .core import Dispatcher, Lifecycle, Plugin, Runner, ScenarioDiscoverer
from .core._config_loader import ConfigFileLoader
from .core._scenario_finder import ScenarioFileFinder
from .core._scenario_finder._file_filters import AnyFilter, DunderFilter, ExtFilter, HiddenFilter
from .core._scenario_loader import ScenarioAssertRewriterLoader
from .plugins.deferrer import defer
from .plugins.skipper import only, skip

__version__ = version
__all__ = ("Scenario", "Interface", "Runner", "run", "only", "skip", "params",
           "context", "defer", "Config",)


def run(*, plugins: Optional[List[Plugin]] = None) -> None:
    if plugins is not None:
        raise DeprecationWarning("Argument 'plugins' is deprecated, "
                                 "declare plugins in config (vedro.cfg.py)")

    cwd = os.getcwd()
    if cwd not in sys.path:
        sys.path.insert(0, cwd)

    finder = ScenarioFileFinder(
        file_filter=AnyFilter([
            HiddenFilter(),
            DunderFilter(),
            ExtFilter(only=["py"]),
        ]),
        dir_filter=AnyFilter([
            HiddenFilter(),
            DunderFilter(),
        ])
    )
    discoverer = ScenarioDiscoverer(finder, ScenarioAssertRewriterLoader())
    dispatcher = Dispatcher()
    runner = Runner(dispatcher, (KeyboardInterrupt, SystemExit, CancelledError,))
    lifecycle = Lifecycle(dispatcher, discoverer, runner, ConfigFileLoader(Config))

    asyncio.run(lifecycle.start())



from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when
from rich.console import Style

from vedro.core import Dispatcher, ScenarioResult
from vedro.events import ArgParsedEvent, ScenarioFailedEvent, ScenarioPassedEvent
from vedro.plugins.director import DirectorPlugin, RichReporterPlugin
from vedro.plugins.director.rich.test_utils import (
    chose_reporter,
    console_,
    director,
    dispatcher,
    make_parsed_args,
    make_vscenario,
    reporter,
)

__all__ = ("dispatcher", "director", "reporter", "console_",)


@pytest.mark.asyncio
async def test_rich_reporter(*, dispatcher: Dispatcher, director: DirectorPlugin,
                             reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        event = ArgParsedEvent(make_parsed_args(reruns=1))
        await dispatcher.fire(event)

        vscenario = make_vscenario()
        scenario_result = ScenarioResult(vscenario, rerun=0).mark_failed()
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == []


@pytest.mark.asyncio
async def test_rich_reporter_failed(*, dispatcher: Dispatcher, director: DirectorPlugin,
                                    reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        event = ArgParsedEvent(make_parsed_args(reruns=1))
        await dispatcher.fire(event)

        subject = "<subject>"
        vscenario = make_vscenario(subject=subject)

        scenario_result1 = ScenarioResult(vscenario, rerun=0).mark_failed()
        event1 = ScenarioFailedEvent(scenario_result1)
        await dispatcher.fire(event1)

        scenario_result2 = ScenarioResult(vscenario, rerun=1).mark_failed()
        event2 = ScenarioFailedEvent(scenario_result2)

    with when:
        await dispatcher.fire(event2)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {subject}", style=Style.parse("red")),

            call.out(" │"),
            call.out(" ├─[1/2]", end=""),
            call.out(f" ✗ {subject}", style=Style.parse("red")),

            call.out(" │"),
            call.out(" ├─[2/2]", end=""),
            call.out(f" ✗ {subject}", style=Style.parse("red")),

            call.out(" ")
        ]


@pytest.mark.asyncio
async def test_rich_reporter_passed(*, dispatcher: Dispatcher, director: DirectorPlugin,
                                    reporter: RichReporterPlugin, console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        event = ArgParsedEvent(make_parsed_args(reruns=2))
        await dispatcher.fire(event)

        subject = "<subject>"
        vscenario = make_vscenario(subject=subject)

        scenario_result1 = ScenarioResult(vscenario, rerun=0).mark_failed()
        event1 = ScenarioFailedEvent(scenario_result1)
        await dispatcher.fire(event1)

        scenario_result2 = ScenarioResult(vscenario, rerun=1).mark_passed()
        event2 = ScenarioPassedEvent(scenario_result2)
        await dispatcher.fire(event2)

        scenario_result3 = ScenarioResult(vscenario, rerun=2).mark_passed()
        event3 = ScenarioPassedEvent(scenario_result3)

    with when:
        await dispatcher.fire(event3)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✔ {subject}", style=Style.parse("green")),

            call.out(" │"),
            call.out(" ├─[1/3]", end=""),
            call.out(f" ✗ {subject}", style=Style.parse("red")),

            call.out(" │"),
            call.out(" ├─[2/3]", end=""),
            call.out(f" ✔ {subject}", style=Style.parse("green")),

            call.out(" │"),
            call.out(" ├─[3/3]", end=""),
            call.out(f" ✔ {subject}", style=Style.parse("green")),

            call.out(" ")
        ]



from baby_steps import then, when
from pytest import raises

from vedro.core import Config, ConfigLoader


def test_config_loader():
    with when, raises(Exception) as exc_info:
        ConfigLoader(Config)

    with then:
        assert exc_info.type is TypeError
        assert "Can't instantiate abstract class ConfigLoader" in str(exc_info.value)



from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when
from rich.console import Style

from vedro.core import Dispatcher, ScenarioResult, StepResult
from vedro.events import ArgParsedEvent, ScenarioFailedEvent
from vedro.plugins.director import DirectorPlugin, RichReporterPlugin
from vedro.plugins.director.rich.test_utils import (
    chose_reporter,
    console_,
    director,
    dispatcher,
    make_exc_info,
    make_parsed_args,
    make_scenario_result,
    make_step_result,
    make_vscenario,
    make_vstep,
    reporter,
)

__all__ = ("dispatcher", "reporter", "director", "console_",)


@pytest.mark.asyncio
async def test_rich_reporter_scenario_failed_event_default_verbose(*, dispatcher: Dispatcher,
                                                                   director: DirectorPlugin,
                                                                   reporter: RichReporterPlugin,
                                                                   console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        subject = "<subject>"
        vscenario = make_vscenario(subject=subject)
        scenario_result = ScenarioResult(vscenario).mark_failed()
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {subject}", style=Style.parse("red"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_failed_event_verbose0(*, dispatcher: Dispatcher,
                                                            director: DirectorPlugin,
                                                            reporter: RichReporterPlugin,
                                                            console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        event = ArgParsedEvent(make_parsed_args(verbose=0))
        await dispatcher.fire(event)

        scenario_result = make_scenario_result().mark_failed()
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {scenario_result.scenario.subject}", style=Style.parse("red"))
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_failed_event_verbose1(*, dispatcher: Dispatcher,
                                                            director: DirectorPlugin,
                                                            reporter: RichReporterPlugin,
                                                            console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        event = ArgParsedEvent(make_parsed_args(verbose=1))
        await dispatcher.fire(event)

        subject = "<subject>"
        vscenario = make_vscenario(subject=subject)

        step_name_passed = "<passed step>"
        step_result_passed = StepResult(make_vstep(name=step_name_passed)).mark_passed()
        step_name_failed = "<failed step>"
        step_result_failed = StepResult(make_vstep(name=step_name_failed)).mark_failed()
        step_result = make_step_result()

        scenario_result = make_scenario_result(vscenario, [
            step_result_passed,
            step_result_failed,
            step_result,
        ]).mark_failed()
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {subject}", style=Style.parse("red")),
            call.out(f"    ✔ {step_name_passed}", style=Style.parse("green")),
            call.out(f"    ✗ {step_name_failed}", style=Style.parse("red")),
        ]


@pytest.mark.asyncio
async def test_rich_reporter_scenario_failed_event_verbose2(*, dispatcher: Dispatcher,
                                                            director: DirectorPlugin,
                                                            reporter: RichReporterPlugin,
                                                            console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        event = ArgParsedEvent(make_parsed_args(verbose=2))
        await dispatcher.fire(event)

        step_result_passed = make_step_result().mark_passed()
        exc_info = make_exc_info(AssertionError())
        step_result_failed = make_step_result().mark_failed().set_exc_info(exc_info)
        step_result = make_step_result()

        scenario_result = make_scenario_result(step_results=[
            step_result_passed,
            step_result_failed,
            step_result,
        ]).mark_failed()
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls[:3] == [
            call.out(f" ✗ {scenario_result.scenario.subject}", style=Style.parse("red")),
            call.out(f"    ✔ {step_result_passed.step_name}", style=Style.parse("green")),
            call.out(f"    ✗ {step_result_failed.step_name}", style=Style.parse("red")),
        ]
        name, args, kwargs = console_.mock_calls[3]
        assert "Traceback" in args[0]
        assert kwargs["style"] == Style.parse("yellow")


@pytest.mark.asyncio
async def test_rich_reporter_scenario_failed_event_verbose3(*, dispatcher: Dispatcher,
                                                            director: DirectorPlugin,
                                                            reporter: RichReporterPlugin,
                                                            console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        event = ArgParsedEvent(make_parsed_args(verbose=3))
        await dispatcher.fire(event)

        step_result_passed = make_step_result().mark_passed()
        exc_info = make_exc_info(AssertionError())
        step_result_failed = make_step_result().mark_failed().set_exc_info(exc_info)
        step_result = make_step_result()

        scenario_result = make_scenario_result(step_results=[
            step_result_passed,
            step_result_failed,
            step_result,
        ]).mark_failed()
        scenario_result.set_scope({"key_int": 1, "key_str": "val"})
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls[:3] == [
            call.out(f" ✗ {scenario_result.scenario.subject}", style=Style.parse("red")),
            call.out(f"    ✔ {step_result_passed.step_name}", style=Style.parse("green")),
            call.out(f"    ✗ {step_result_failed.step_name}", style=Style.parse("red")),
        ]

        name, args, kwargs = console_.mock_calls[3]
        assert "Traceback" in args[0]
        assert kwargs["style"] == Style.parse("yellow")

        name, args, kwargs = console_.mock_calls[4]
        assert "Scope" in args[0]
        assert kwargs["style"] == Style(color="blue", bold=True)


@pytest.mark.asyncio
async def test_rich_reporter_scenario_failed_event_without_steps_verbose3(
        *,
        dispatcher: Dispatcher,
        director: DirectorPlugin,
        reporter: RichReporterPlugin,
        console_: Mock):
    with given:
        await chose_reporter(dispatcher, director, reporter)

        event = ArgParsedEvent(make_parsed_args(verbose=3))
        await dispatcher.fire(event)

        scenario_result = make_scenario_result().mark_failed()
        event = ScenarioFailedEvent(scenario_result)

    with when:
        await dispatcher.fire(event)

    with then:
        assert console_.mock_calls == [
            call.out(f" ✗ {scenario_result.scenario.subject}", style=Style.parse("red")),
        ]



from abc import abstractmethod
from typing import Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig

__all__ = ("Reporter",)


class Reporter(Plugin):
    def __init__(self, config: Type[PluginConfig]) -> None:
        super().__init__(config)
        self._dispatcher: Union[Dispatcher, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        self._dispatcher = dispatcher

    @abstractmethod
    def on_chosen(self) -> None:
        raise NotImplementedError()



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from time import monotonic_ns
from typing import Type, cast
from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when

from vedro.core import Dispatcher, Event, Subscriber


@pytest.fixture()
def dispatcher():
    return Dispatcher()


@pytest.fixture()
def event_type() -> Type[Event]:
    name = "CustomEvent_{}".format(monotonic_ns())
    return cast(Type[Event], type(name, (Event,), {}))


def test_dispatcher_register(*, dispatcher: Dispatcher):
    with given:
        subcriber_ = Mock(Subscriber)

    with when:
        res = dispatcher.register(subcriber_)

    with then:
        assert res is None
        assert subcriber_.mock_calls == [call.subscribe(dispatcher)]


def test_dispatcher_listen(*, dispatcher: Dispatcher, event_type: Type[Event]):
    with when:
        res = dispatcher.listen(event_type, lambda e: None)

    with then:
        assert res == dispatcher


@pytest.mark.asyncio
async def test_dispatcher_fire_unknown_event(*, dispatcher: Dispatcher, event_type: Type[Event]):
    with given:
        event = event_type()

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None


@pytest.mark.asyncio
@pytest.mark.parametrize("mock_factory", (Mock, AsyncMock))
async def test_dispatcher_fire(mock_factory, *, dispatcher: Dispatcher, event_type: Type[Event]):
    with given:
        subscribe_ = mock_factory()
        dispatcher.listen(event_type, subscribe_)
        event = event_type()

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert subscribe_.mock_calls == [call(event)]


@pytest.mark.asyncio
async def test_dispatcher_fire_twice(*, dispatcher: Dispatcher, event_type: Type[Event]):
    with given:
        subscribe_ = Mock()
        dispatcher.listen(event_type, subscribe_)
        event = event_type()
        await dispatcher.fire(event)

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert subscribe_.mock_calls == [call(event), call(event)]


@pytest.mark.asyncio
async def test_dispatcher_fire_default_order(*, dispatcher: Dispatcher, event_type: Type[Event]):
    with given:
        manager_ = Mock()
        subscribe1_, subscribe2_ = Mock(), Mock()
        manager_.attach_mock(subscribe1_, "subscribe1_")
        manager_.attach_mock(subscribe2_, "subscribe2_")

        dispatcher.listen(event_type, subscribe2_)
        dispatcher.listen(event_type, subscribe1_)
        event = event_type()

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert manager_.mock_calls == [call.subscribe2_(event), call.subscribe1_(event)]


@pytest.mark.asyncio
async def test_dispatcher_fire_priority_order(*, dispatcher: Dispatcher, event_type: Type[Event]):
    with given:
        manager_ = Mock()
        subscribe1_, subscribe2_ = Mock(), Mock()
        manager_.attach_mock(subscribe1_, "subscribe1_")
        manager_.attach_mock(subscribe2_, "subscribe2_")

        dispatcher.listen(event_type, subscribe2_, priority=1)
        dispatcher.listen(event_type, subscribe1_)
        event = event_type()

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert manager_.mock_calls == [call.subscribe1_(event), call.subscribe2_(event)]


@pytest.mark.asyncio
async def test_dispatcher_fire_listen_default_order(*, dispatcher: Dispatcher,
                                                    event_type: Type[Event]):
    with given:
        subscribe3_ = Mock()

        def handler(e):
            dispatcher.listen(event_type, subscribe3_)
        subscribe1_, subscribe2_ = Mock(), Mock(side_effect=handler)

        manager_ = Mock()
        manager_.attach_mock(subscribe1_, "subscribe1_")
        manager_.attach_mock(subscribe2_, "subscribe2_")
        manager_.attach_mock(subscribe3_, "subscribe3_")

        dispatcher.listen(event_type, subscribe2_)
        dispatcher.listen(event_type, subscribe1_)
        event = event_type()

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert manager_.mock_calls == [
            call.subscribe2_(event),
            call.subscribe1_(event),
            call.subscribe3_(event),
        ]


@pytest.mark.asyncio
async def test_dispatcher_fire_listen_priority_order(*, dispatcher: Dispatcher,
                                                     event_type: Type[Event]):
    with given:
        subscribe3_ = Mock()

        def handler(e):
            dispatcher.listen(event_type, subscribe3_)
        subscribe1_, subscribe2_ = Mock(), Mock(side_effect=handler)

        manager_ = Mock()
        manager_.attach_mock(subscribe1_, "subscribe1_")
        manager_.attach_mock(subscribe2_, "subscribe2_")
        manager_.attach_mock(subscribe3_, "subscribe3_")

        dispatcher.listen(event_type, subscribe2_)
        dispatcher.listen(event_type, subscribe1_, priority=1)
        event = event_type()

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert manager_.mock_calls == [
            call.subscribe2_(event),
            call.subscribe3_(event),
            call.subscribe1_(event),
        ]



from baby_steps import given, then, when

from vedro.plugins.director import Director, DirectorInitEvent, DirectorPlugin


def test_director_init_event_prop():
    with given:
        director = DirectorPlugin(Director)
        init_event = DirectorInitEvent(director)

    with when:
        res = init_event.director

    with then:
        assert res == director


def test_director_init_event_repr():
    with given:
        director = DirectorPlugin(Director)
        init_event = DirectorInitEvent(director)

    with when:
        res = repr(init_event)

    with then:
        assert res == f"DirectorInitEvent({director!r})"



from collections import deque
from time import monotonic_ns
from unittest.mock import Mock

import pytest
from baby_steps import given, then, when

from vedro.core import Dispatcher, Plugin, ScenarioResult, StepResult, VirtualScenario, VirtualStep
from vedro.events import (
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    StepFailedEvent,
    StepPassedEvent,
)
from vedro.plugins.artifacted import Artifacted, ArtifactedPlugin, MemoryArtifact


@pytest.fixture()
def dispatcher() -> Dispatcher:
    return Dispatcher()


@pytest.fixture()
def scenario_artifacts() -> deque:
    return deque()


@pytest.fixture()
def step_artifacts() -> deque:
    return deque()


@pytest.fixture()
def artifacted(scenario_artifacts: deque, step_artifacts: deque) -> ArtifactedPlugin:
    return ArtifactedPlugin(Artifacted,
                            scenario_artifacts=scenario_artifacts, step_artifacts=step_artifacts)


def create_artifact() -> MemoryArtifact:
    return MemoryArtifact(f"test-{monotonic_ns()}", "text/plain", b"")


def test_artifacted_plugin():
    with when:
        plugin = ArtifactedPlugin(Artifacted)

    with then:
        assert isinstance(plugin, Plugin)


@pytest.mark.asyncio
async def test_artifacted_scenario_run_event(*, dispatcher: Dispatcher,
                                             artifacted: ArtifactedPlugin,
                                             scenario_artifacts: deque, step_artifacts: deque):
    with given:
        artifacted.subscribe(dispatcher)

        scenario_result = ScenarioResult(Mock(VirtualScenario))
        event = ScenarioRunEvent(scenario_result)

        scenario_artifacts.append(create_artifact())
        step_artifacts.append(create_artifact())

    with when:
        await dispatcher.fire(event)

    with then:
        assert len(scenario_artifacts) == 0
        assert len(step_artifacts) == 0


@pytest.mark.asyncio
@pytest.mark.parametrize("event_class", [ScenarioPassedEvent, ScenarioFailedEvent])
async def test_artifacted_scenario_end_event(event_class, *, dispatcher: Dispatcher,
                                             artifacted: ArtifactedPlugin,
                                             scenario_artifacts: deque):
    with given:
        artifacted.subscribe(dispatcher)
        scenario_result = ScenarioResult(Mock(VirtualScenario))
        event = event_class(scenario_result)

        artifact1 = create_artifact()
        scenario_artifacts.append(artifact1)

        artifact2 = create_artifact()
        scenario_artifacts.append(artifact2)

    with when:
        await dispatcher.fire(event)

    with then:
        assert scenario_result.artifacts == [artifact1, artifact2]


@pytest.mark.asyncio
@pytest.mark.parametrize("event_class", [StepPassedEvent, StepFailedEvent])
async def test_artifacted_step_end_event(event_class, *, dispatcher: Dispatcher,
                                         artifacted: ArtifactedPlugin,
                                         step_artifacts: deque):
    with given:
        artifacted.subscribe(dispatcher)
        step_result = StepResult(Mock(VirtualStep))
        event = event_class(step_result)

        artifact1 = create_artifact()
        step_artifacts.append(artifact1)

        artifact2 = create_artifact()
        step_artifacts.append(artifact2)

    with when:
        await dispatcher.fire(event)

    with then:
        assert step_result.artifacts == [artifact1, artifact2]



from typing import Any, Type

import cabina
from cabina import MetaBase
from cabina.errors import ConfigError

__all__ = ("Config", "Section", "ConfigType",)


class _MetaBase(MetaBase):
    def __call__(cls, *args: Any, **kwargs: Any) -> None:
        try:
            super().__call__(*args, **kwargs)
        except ConfigError:
            # backward compatibility
            raise DeprecationWarning("Declare plugin in config (vedro.cfg.py)")


class Section(cabina.Section, metaclass=_MetaBase):
    pass


class Config(cabina.Config, Section):
    class Plugins(Section):
        pass


ConfigType = Type[Config]



from collections import deque
from typing import Deque, Type, Union

from vedro.core import Artifact, Dispatcher, Plugin, PluginConfig
from vedro.events import (
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    StepFailedEvent,
    StepPassedEvent,
)

__all__ = ("Artifacted", "ArtifactedPlugin",
           "attach_artifact", "attach_step_artifact", "attach_scenario_artifact",)


_scenario_artifacts: Deque[Artifact] = deque()
_step_artifacts: Deque[Artifact] = deque()


def attach_scenario_artifact(artifact: Artifact) -> None:
    _scenario_artifacts.append(artifact)


def attach_step_artifact(artifact: Artifact) -> None:
    _step_artifacts.append(artifact)


def attach_artifact(artifact: Artifact) -> None:
    attach_step_artifact(artifact)


class ArtifactedPlugin(Plugin):
    def __init__(self, config: Type["Artifacted"], *,
                 scenario_artifacts: Deque[Artifact] = _scenario_artifacts,
                 step_artifacts: Deque[Artifact] = _step_artifacts) -> None:
        super().__init__(config)
        self._scenario_artifacts = scenario_artifacts
        self._step_artifacts = step_artifacts

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(StepPassedEvent, self.on_step_end) \
                  .listen(StepFailedEvent, self.on_step_end) \
                  .listen(ScenarioPassedEvent, self.on_scenario_end) \
                  .listen(ScenarioFailedEvent, self.on_scenario_end)

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        self._scenario_artifacts.clear()
        self._step_artifacts.clear()

    async def on_step_end(self, event: Union[StepPassedEvent, StepFailedEvent]) -> None:
        while len(self._step_artifacts) > 0:
            artifact = self._step_artifacts.popleft()
            event.step_result.attach(artifact)

    async def on_scenario_end(self,
                              event: Union[ScenarioPassedEvent, ScenarioFailedEvent]) -> None:
        while len(self._scenario_artifacts) > 0:
            artifact = self._scenario_artifacts.popleft()
            event.scenario_result.attach(artifact)


class Artifacted(PluginConfig):
    plugin = ArtifactedPlugin



from argparse import ArgumentParser

import pytest
from baby_steps import given, then, when

from vedro.core import Dispatcher
from vedro.events import ArgParseEvent
from vedro.plugins.seeder import Seeder, SeederPlugin


@pytest.fixture()
def dispatcher() -> Dispatcher:
    return Dispatcher()


@pytest.mark.asyncio
async def test_seeder_plugin(*, dispatcher: Dispatcher):
    with given:
        seeder = SeederPlugin(Seeder)
        seeder.subscribe(dispatcher)
        event = ArgParseEvent(ArgumentParser())

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None



from argparse import ArgumentParser, Namespace
from pathlib import Path
from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when

from vedro.core import Config, Dispatcher
from vedro.events import ArgParseEvent, ConfigLoadedEvent
from vedro.plugins.director import Director, DirectorInitEvent, DirectorPlugin, Reporter


@pytest.fixture()
def dispatcher() -> Dispatcher:
    return Dispatcher()


@pytest.fixture()
def director(dispatcher: Dispatcher) -> DirectorPlugin:
    director = DirectorPlugin(Director)
    director.subscribe(dispatcher)
    return director


@pytest.mark.asyncio
async def test_director_plugin_init_event(*, director: DirectorPlugin, dispatcher: Dispatcher):
    with given:
        callback_ = Mock()
        dispatcher.listen(DirectorInitEvent, callback_)

        event = ConfigLoadedEvent(Path(), Config)

    with when:
        await dispatcher.fire(event)

    with then:
        assert callback_.mock_calls == [call(DirectorInitEvent(director))]


@pytest.mark.asyncio
async def test_director_plugin_with_reporters_default(*, director: DirectorPlugin,
                                                      dispatcher: Dispatcher):
    with given:
        reporter1_ = Mock(Reporter)
        dispatcher.listen(DirectorInitEvent,
                          lambda e: e.director.register("reporter1", reporter1_))

        reporter2_ = Mock(Reporter)
        dispatcher.listen(DirectorInitEvent,
                          lambda e: e.director.register("reporter2", reporter2_))

        await dispatcher.fire(ConfigLoadedEvent(Path(), Config))
        event = ArgParseEvent(ArgumentParser())

    with when:
        await dispatcher.fire(event)

    with then:
        assert reporter1_.mock_calls == [call.on_chosen()]
        assert reporter2_.mock_calls == []


@pytest.mark.asyncio
async def test_director_plugin_with_reporters_arg(*, director: DirectorPlugin,
                                                  dispatcher: Dispatcher):
    with given:
        reporter1_ = Mock(Reporter)
        dispatcher.listen(DirectorInitEvent,
                          lambda e: e.director.register("reporter1", reporter1_))

        reporter2_ = Mock(Reporter)
        dispatcher.listen(DirectorInitEvent,
                          lambda e: e.director.register("reporter2", reporter2_))

        await dispatcher.fire(ConfigLoadedEvent(Path(), Config))

        args = Namespace(reporters=["reporter2", "reporter1"])
        arg_parser = Mock(ArgumentParser, parse_known_args=Mock(return_value=(args, [])))
        event = ArgParseEvent(arg_parser)

    with when:
        await dispatcher.fire(event)

    with then:
        assert reporter1_.mock_calls == [call.on_chosen()]
        assert reporter2_.mock_calls == [call.on_chosen()]



from typing import Type

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import ArgParsedEvent, ArgParseEvent, ScenarioFailedEvent, ScenarioRunEvent

__all__ = ("Interrupter", "InterrupterPlugin",)


class _Interrupted(Exception):
    pass


class InterrupterPlugin(Plugin):
    def __init__(self, config: Type["Interrupter"]) -> None:
        super().__init__(config)
        self._fail_fast: bool = False
        self._failed: int = 0

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(ScenarioFailedEvent, self.on_scenario_failed)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("--fail-fast",
                                      action='store_true',
                                      default=self._fail_fast,
                                      help="Stop after first failed scenario")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._fail_fast = event.args.fail_fast

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        if self._fail_fast and self._failed > 0:
            raise _Interrupted()

    def on_scenario_failed(self, event: ScenarioFailedEvent) -> None:
        if self._fail_fast:
            self._failed += 1


class Interrupter(PluginConfig):
    plugin = InterrupterPlugin



from types import MethodType
from unittest.mock import Mock

import pytest
from baby_steps import given, then, when

from vedro.core import ExcInfo, MemoryArtifact, StepResult, StepStatus, VirtualStep


@pytest.fixture()
def method_():
    return Mock(MethodType)


@pytest.fixture()
def virtual_step(method_: MethodType):
    return VirtualStep(method_)


def test_step_result(*, method_: MethodType, virtual_step: VirtualStep):
    with given:
        method_.__name__ = "<name>"

    with when:
        step_result = StepResult(virtual_step)

    with then:
        assert step_result.step_name == method_.__name__
        assert step_result.is_passed() is False
        assert step_result.is_failed() is False
        assert step_result.started_at is None
        assert step_result.ended_at is None
        assert step_result.exc_info is None
        assert step_result.status == StepStatus.PENDING
        assert repr(step_result) == f"StepResult({virtual_step!r})"


def test_step_result_mark_passed(*, virtual_step: VirtualStep):
    with given:
        step_result = StepResult(virtual_step)

    with when:
        res = step_result.mark_passed()

    with then:
        assert res == step_result
        assert step_result.is_passed() is True


def test_step_result_mark_failed(*, virtual_step: VirtualStep):
    with given:
        step_result = StepResult(virtual_step)

    with when:
        res = step_result.mark_failed()

    with then:
        assert res == step_result
        assert step_result.is_failed() is True


def test_step_result_set_started_at(*, virtual_step: VirtualStep):
    with given:
        step_result = StepResult(virtual_step)
        started_at = 1.0

    with when:
        res = step_result.set_started_at(started_at)

    with then:
        assert res == step_result
        assert step_result.started_at == started_at
        assert step_result.elapsed == 0.0


def test_step_result_set_ended_at(*, virtual_step: VirtualStep):
    with given:
        step_result = StepResult(virtual_step)
        ended_at = 1.0

    with when:
        res = step_result.set_ended_at(ended_at)

    with then:
        assert res == step_result
        assert step_result.ended_at == ended_at
        assert step_result.elapsed == 0.0


def test_step_result_elapsed(*, virtual_step: VirtualStep):
    with given:
        step_result = StepResult(virtual_step)
        started_at = 3.0
        step_result.set_started_at(started_at)
        ended_at = 1.0
        step_result.set_ended_at(ended_at)

    with when:
        res = step_result.elapsed

    with then:
        assert res == ended_at - started_at


def test_step_result_set_exc_info(*, virtual_step: VirtualStep):
    with given:
        step_result = StepResult(virtual_step)
        exc_info = ExcInfo(AssertionError, AssertionError(), None)

    with when:
        res = step_result.set_exc_info(exc_info)

    with then:
        assert res == step_result
        assert step_result.exc_info == exc_info


def test_step_result_eq(*, virtual_step: VirtualStep):
    with given:
        step_result1 = StepResult(virtual_step)
        step_result2 = StepResult(virtual_step)

    with when:
        res = step_result1 == step_result2

    with then:
        assert res is True


def test_step_result_not_eq():
    with given:
        virtual_step1 = VirtualStep(Mock(MethodType))
        step_result1 = StepResult(virtual_step1)
        virtual_step2 = VirtualStep(Mock(MethodType))
        step_result2 = StepResult(virtual_step2)

    with when:
        res = step_result1 == step_result2

    with then:
        assert res is False


def test_step_result_attach_artifact(*, virtual_step: VirtualStep):
    with given:
        step_result = StepResult(virtual_step)
        artifact = MemoryArtifact("name", "text/plain", b"")

    with when:
        res = step_result.attach(artifact)

    with then:
        assert res is None


def test_step_result_get_artifacts(*, virtual_step: VirtualStep):
    with given:
        step_result = StepResult(virtual_step)

        artifact1 = MemoryArtifact("name1", "text/plain", b"")
        step_result.attach(artifact1)

        artifact2 = MemoryArtifact("name2", "text/plain", b"")
        step_result.attach(artifact2)

    with when:
        artifacts = step_result.artifacts

    with then:
        assert artifacts == [artifact1, artifact2]



import os
from typing import List, Optional, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig, VirtualScenario
from vedro.events import ArgParsedEvent, ArgParseEvent, StartupEvent

__all__ = ("Skipper", "SkipperPlugin",)


class _CompositePath:
    def __init__(self, file_path: str, cls_name: Optional[str], tmpl_idx: Optional[int]) -> None:
        self.file_path = file_path
        self.cls_name = cls_name
        self.tmpl_idx = tmpl_idx


class SkipperPlugin(Plugin):
    def __init__(self, config: Type["Skipper"]) -> None:
        super().__init__(config)
        self._subject: Union[str, None] = None
        self._specified: List[_CompositePath] = []
        self._ignored: List[_CompositePath] = []

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("file_or_dir", nargs="*", default=["."],
                                      help="Select scenarios in a given file or directory")
        event.arg_parser.add_argument("-i", "--ignore", nargs="+", default=[],
                                      help="Skip scenarios in a given file or directory")
        event.arg_parser.add_argument("--subject", help="Select scenarios with a given subject")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._subject = event.args.subject

        for file_or_dir in event.args.file_or_dir:
            composite_path = self._get_composite_path(file_or_dir)
            path = composite_path.file_path
            assert os.path.isdir(path) or os.path.isfile(path), f"{path!r} does not exist"
            self._specified.append(composite_path)

        for file_or_dir in event.args.ignore:
            composite_path = self._get_composite_path(file_or_dir)
            path = composite_path.file_path
            assert os.path.isdir(path) or os.path.isfile(path), f"{path!r} does not exist"
            self._ignored.append(composite_path)

    def _get_composite_path(self, file_or_dir: str) -> _CompositePath:
        head, tail = os.path.split(file_or_dir)
        file_name, *other = tail.split("::")
        cls_name, *other = "".join(other).split("#")
        tmpl_idx = "".join(other)
        return _CompositePath(
            file_path=self._normalize_path(os.path.join(head, file_name)),
            cls_name=cls_name if len(cls_name) > 0 else None,
            tmpl_idx=int(tmpl_idx) if tmpl_idx.isnumeric() else None,
        )

    def _normalize_path(self, file_or_dir: str) -> str:
        path = os.path.normpath(file_or_dir)
        if os.path.isabs(path):
            return path
        if os.path.commonpath(["scenarios", path]) != "scenarios":
            path = os.path.join("scenarios", path)
        return os.path.abspath(path)

    def _is_scenario_skipped(self, scenario: VirtualScenario) -> bool:
        if getattr(scenario._orig_scenario, "__vedro__skipped__", False):
            return True
        template = getattr(scenario._orig_scenario, "__vedro__template__", None)
        if getattr(template, "__vedro__skipped__", False):
            return True
        if self._subject and scenario.subject != self._subject:
            return True
        if not self._is_scenario_specified(scenario):
            return True
        if self._is_scenario_ignored(scenario):
            return True
        return False

    def _is_scenario_special(self, scenario: VirtualScenario) -> bool:
        if getattr(scenario._orig_scenario, "__vedro__only__", False):
            return True
        template = getattr(scenario._orig_scenario, "__vedro__template__", None)
        if getattr(template, "__vedro__only__", False):
            return True
        return False

    def _is_match_scenario(self, path: _CompositePath, scenario: VirtualScenario) -> bool:
        if os.path.commonpath([path.file_path, scenario.path]) != path.file_path:
            return False

        cls_name = scenario._orig_scenario.__name__
        # Fix matching logic (startswith)
        if (path.cls_name is not None) and (not cls_name.startswith(path.cls_name)):
            return False

        tmpl_idx = scenario.template_index
        if (path.tmpl_idx is not None) and (path.tmpl_idx != tmpl_idx):
            return False

        return True

    def _is_scenario_specified(self, scenario: VirtualScenario) -> bool:
        for path in self._specified:
            if self._is_match_scenario(path, scenario):
                return True
        return False

    def _is_scenario_ignored(self, scenario: VirtualScenario) -> bool:
        for path in self._ignored:
            if self._is_match_scenario(path, scenario):
                return True
        return False

    def on_startup(self, event: StartupEvent) -> None:
        special_scenarios = set()
        for scenario in event.scenarios:
            if self._is_scenario_skipped(scenario):
                scenario.skip()
            elif self._is_scenario_special(scenario):
                special_scenarios.add(scenario.unique_id)

        if len(special_scenarios) > 0:
            for scenario in event.scenarios:
                if scenario.unique_id not in special_scenarios:
                    scenario.skip()


class Skipper(PluginConfig):
    plugin = SkipperPlugin



from pathlib import Path

import pytest

from vedro.core._scenario_finder._file_filters import HiddenFilter


@pytest.mark.parametrize(("value", "expected"), [
    (".DS_Store", True),
    ("/tmp/.DS_Store", True),
    #
    ("main.py", False),
    ("/tmp/main.py", False),
])
def test_hidden_file_filter(value: str, expected: bool):
    hidden_filter = HiddenFilter()

    filtered = hidden_filter.filter(Path(value))
    assert filtered is expected



from baby_steps import then, when
from pytest import raises

from vedro.core import PluginConfig
from vedro.plugins.director import Reporter


def test_reporter():
    with when, raises(BaseException) as exception:
        Reporter(PluginConfig)

    with then:
        assert exception.type is TypeError



from baby_steps import then, when
from pytest import raises

from vedro.core import ScenarioLoader


def test_scenario_loader():
    with when, raises(Exception) as exc_info:
        ScenarioLoader()

    with then:
        assert exc_info.type is TypeError
        assert "Can't instantiate abstract class ScenarioLoader" in str(exc_info.value)



from ._director import Director, DirectorPlugin
from ._director_init_event import DirectorInitEvent
from ._reporter import Reporter
from .pycharm import PyCharmReporter, PyCharmReporterPlugin
from .rich import RichReporter, RichReporterPlugin
from .silent import SilentReporter, SilentReporterPlugin

__all__ = ("Director", "DirectorPlugin", "DirectorInitEvent", "Reporter",
           "RichReporter", "RichReporterPlugin",
           "SilentReporter", "SilentReporterPlugin",
           "PyCharmReporter", "PyCharmReporterPlugin",)



from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when

from vedro.core import Dispatcher
from vedro.plugins.director import (
    DirectorInitEvent,
    DirectorPlugin,
    PyCharmReporter,
    PyCharmReporterPlugin,
)


@pytest.mark.asyncio
async def test_pycharm_reporter_subscribe():
    with given:
        dispatcher = Dispatcher()
        director_ = Mock(DirectorPlugin)

        reporter = PyCharmReporterPlugin(PyCharmReporter)
        reporter.subscribe(dispatcher)

    with when:
        await dispatcher.fire(DirectorInitEvent(director_))

    with then:
        assert director_.mock_calls == [
            call.register("pycharm", reporter)
        ]


def test_pycharm_reporter_on_chosen():
    with given:
        dispatcher = Dispatcher()

        reporter = PyCharmReporterPlugin(PyCharmReporter)
        reporter.subscribe(dispatcher)

    with when:
        res = reporter.on_chosen()

    with then:
        assert res is None



from typing import Type

from vedro.core import Dispatcher, PluginConfig

from .._director_init_event import DirectorInitEvent
from .._reporter import Reporter

__all__ = ("SilentReporter", "SilentReporterPlugin",)


class SilentReporterPlugin(Reporter):
    def __init__(self, config: Type["SilentReporter"]) -> None:
        super().__init__(config)

    def subscribe(self, dispatcher: Dispatcher) -> None:
        super().subscribe(dispatcher)
        dispatcher.listen(DirectorInitEvent, lambda e: e.director.register("silent", self))

    def on_chosen(self) -> None:
        pass


class SilentReporter(PluginConfig):
    plugin = SilentReporterPlugin



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from argparse import Namespace
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
from baby_steps import given, then, when

from vedro.core import (
    Config,
    ConfigFileLoader,
    ConfigLoader,
    Dispatcher,
    Lifecycle,
    Report,
    Runner,
    ScenarioDiscoverer,
)


@pytest.fixture()
def dispatcher_():
    return AsyncMock(Dispatcher)


@pytest.fixture()
def discoverer_():
    return Mock(ScenarioDiscoverer)


@pytest.fixture()
def config_loader():
    return ConfigFileLoader(Config)


@pytest.fixture()
def runner_(dispatcher_: Dispatcher):
    report = Report()
    return Mock(Runner(dispatcher_), run=AsyncMock(return_value=report))


@pytest.mark.asyncio
async def test_lifecycle_register_start(*, dispatcher_: Dispatcher,
                                        discoverer_: ScenarioDiscoverer,
                                        runner_: Runner,
                                        config_loader: ConfigLoader):
    with given:
        scenarios = []
        reruns = 2
        discoverer_.discover = AsyncMock(return_value=scenarios)

        lifecycle = Lifecycle(dispatcher_, discoverer_, runner_, config_loader)
        namespace = Namespace(reruns=reruns)

    with when, patch("argparse.ArgumentParser.parse_args", return_value=namespace):
        report = await lifecycle.start()

    with then:
        assert isinstance(report, Report)
        assert discoverer_.mock_calls == [
            call.discover(Path("scenarios"))
        ]
        assert runner_.mock_calls == [
            call.run(scenarios, reruns=reruns),
        ]
        assert len(dispatcher_.mock_calls) > 0


def test_lifecycle_repr(*, dispatcher_: Dispatcher,
                        discoverer_: ScenarioDiscoverer,
                        runner_: Runner,
                        config_loader: ConfigLoader):
    with when:
        lifecycle = Lifecycle(dispatcher_, discoverer_, runner_, config_loader)

    with then:
        assert repr(lifecycle) == (
            f"Lifecycle({dispatcher_!r}, {discoverer_!r}, {runner_!r}, {config_loader!r})"
        )



import os
from hashlib import blake2b
from pathlib import Path
from types import MethodType
from typing import Type
from unittest.mock import Mock

import pytest
from baby_steps import given, then, when

from vedro import Scenario
from vedro.core import VirtualScenario, VirtualStep


@pytest.fixture()
def scenario_():
    scenario = Mock(Scenario)
    scenario.__file__ = os.getcwd() + "/scenarios/scenario.py"
    scenario.__module__ = "scenarios.scenario"
    scenario.__qualname__ = "Scenario"
    return scenario


@pytest.fixture()
def method_():
    return Mock(MethodType)


def test_virtual_scenario_steps(*, scenario_: Type[Scenario], method_: MethodType):
    with given:
        expected_steps = [VirtualStep(method_), VirtualStep(method_)]
        virtual_scenario = VirtualScenario(scenario_, expected_steps)

    with when:
        actual_steps = virtual_scenario.steps

    with then:
        assert actual_steps == expected_steps


def test_virtual_scenario_unique_id(*, scenario_: Type[Scenario]):
    with given:
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        unique_id = virtual_scenario.unique_id

    with then:
        unique_name = f"{scenario_.__file__}::{scenario_.__qualname__}"
        assert unique_id == blake2b(unique_name.encode(), digest_size=20).hexdigest()


def test_virtual_scenario_path(*, scenario_: Type[Scenario]):
    with given:
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        path = virtual_scenario.path

    with then:
        assert path == Path(scenario_.__file__)


def test_virtual_scenario_subject(*, scenario_: Type[Scenario]):
    with given:
        scenario_.subject = "<subject>"
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        subject = virtual_scenario.subject

    with then:
        assert subject == scenario_.subject


def test_virtual_scenario_without_subject(*, scenario_: Type[Scenario]):
    with given:
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        subject = virtual_scenario.subject

    with then:
        assert subject == Path(scenario_.__file__).stem


def test_virtual_scenario_namespace(*, scenario_: Type[Scenario]):
    with given:
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        namespace = virtual_scenario.namespace

    with then:
        assert namespace == ""


def test_virtual_scenario_with_subject(*, scenario_: Type[Scenario]):
    with given:
        scenario_.subject = "<subject>"
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        subject = virtual_scenario.subject

    with then:
        assert subject == scenario_.subject


def test_virtual_scenario_is_skipped(*, scenario_: Type[Scenario]):
    with given:
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        is_skipped = virtual_scenario.is_skipped()

    with then:
        assert is_skipped is False


def test_virtual_scenario_skip(*, scenario_: Type[Scenario]):
    with given:
        virtual_scenario = VirtualScenario(scenario_, [])

    with when:
        res = virtual_scenario.skip()

    with then:
        assert res is None
        assert virtual_scenario.is_skipped() is True


def test_virtual_scenario_repr(*, scenario_: Type[Scenario], method_: MethodType):
    with given:
        steps = [VirtualStep(method_), VirtualStep(method_)]
        virtual_scenario = VirtualScenario(scenario_, steps)

    with when:
        res = repr(virtual_scenario)

    with then:
        assert res == f"VirtualScenario({scenario_!r}, {steps!r})"


def test_virtual_scenario_eq_without_steps(*, scenario_: Type[Scenario]):
    with given:
        virtual_scenario1 = VirtualScenario(scenario_, [])
        virtual_scenario2 = VirtualScenario(scenario_, [])

    with when:
        res = virtual_scenario1 == virtual_scenario2

    with then:
        assert res is True


def test_virtual_scenario_not_eq_without_steps(*, scenario_: Type[Scenario]):
    with given:
        another_scenario_ = Mock(Scenario)
        another_scenario_.__file__ = scenario_.__file__
        another_scenario_.__qualname__ = scenario_.__qualname__
        virtual_scenario1 = VirtualScenario(scenario_, [])
        virtual_scenario2 = VirtualScenario(another_scenario_, [])

    with when:
        res = virtual_scenario1 == virtual_scenario2

    with then:
        assert res is False


def test_virtual_scenario_eq_with_steps(*, scenario_: Type[Scenario], method_: MethodType):
    with given:
        steps = [VirtualStep(method_), VirtualStep(method_)]
        virtual_scenario1 = VirtualScenario(scenario_, steps)
        virtual_scenario2 = VirtualScenario(scenario_, steps)

    with when:
        res = virtual_scenario1 == virtual_scenario2

    with then:
        assert res is True


def test_virtual_scenario_not_eq_with_steps(*, scenario_: Type[Scenario], method_: MethodType):
    with given:
        steps = [VirtualStep(method_), VirtualStep(method_)]
        virtual_scenario1 = VirtualScenario(scenario_, steps)

        another_method_ = Mock(MethodType)
        another_steps = [VirtualStep(another_method_), VirtualStep(another_method_)]
        virtual_scenario2 = VirtualScenario(scenario_, another_steps)

    with when:
        res = virtual_scenario1 == virtual_scenario2

    with then:
        assert res is False



from abc import ABC, abstractmethod
from pathlib import Path

from ._config_type import ConfigType

__all__ = ("ConfigLoader",)


class ConfigLoader(ABC):
    def __init__(self, default_config: ConfigType) -> None:
        self._default_config = default_config

    @abstractmethod
    async def load(self, path: Path) -> ConfigType:
        raise NotImplementedError()



from baby_steps import given, then, when
from rich.console import Console

from vedro.plugins.director.rich.utils import make_console


def test_make_console():
    with when:
        res = make_console()

    with then:
        assert isinstance(res, Console)


def test_make_console_custom_options():
    with given:
        width, height = 1, 2

    with when:
        res = make_console(width=width, height=height)

    with then:
        assert isinstance(res, Console)
        assert res.size == (width, height)



import argparse
import sys
from typing import Any, Optional

__all__ = ("ArgumentParser",)


class ArgumentParser(argparse.ArgumentParser):
    _subparser_action: Optional[argparse.Action] = None
    _subparser_default: Optional[str] = None

    def add_subparsers(self, **kwargs: Any) -> Any:
        action = super().add_subparsers(**kwargs)
        self._subparser_action = action
        return action

    def get_subparsers(self) -> Any:
        return self._subparser_action

    def set_default_subparser(self, name: str) -> None:
        self._subparser_default = name

    def parse_known_args(self, args: Any = None, namespace: Any = None) -> Any:
        if args is None:
            args = sys.argv[1:]
        else:
            args = list(args)

        if self._subparser_default and self._subparser_action and self._subparser_action.choices:
            arguments = set(args)
            commands = set(self._subparser_action.choices)
            help_args = {"-h", "--help"}
            if not help_args.intersection(arguments) and not commands.intersection(arguments):
                args.insert(0, self._subparser_default)

        return super().parse_known_args(args, namespace)



from pathlib import Path

import pytest

from vedro.core._scenario_finder._file_filters import DunderFilter


@pytest.mark.parametrize(("value", "expected"), [
    ("__pycache__", True),
    ("/tmp/__pycache__", True),
    ("__init__.py", True),
    ("/tmp/__init__.py", True),
    ("/tmp/__cache__.tar.gz", True),
    #
    ("/tmp", False),
    ("/tmp/main.py", False),
    ("/tmp/__pycache", False),
    ("main.py", False),
    ("__init.py", False),
])
def test_dunder_file_filter(value: str, expected: bool):
    dunder_filter = DunderFilter()

    filtered = dunder_filter.filter(Path(value))
    assert filtered is expected



from typing import cast
from unittest.mock import Mock

import pytest
from baby_steps import given, then, when
from pytest import raises

from vedro.core import Dispatcher, Report, ScenarioResult, VirtualScenario
from vedro.events import CleanupEvent
from vedro.plugins.terminator import Terminator, TerminatorPlugin


@pytest.fixture()
def dispatcher() -> Dispatcher:
    return Dispatcher()


@pytest.fixture()
def terminator(dispatcher: Dispatcher):
    terminator = TerminatorPlugin(Terminator)
    terminator.subscribe(dispatcher)
    return terminator


def make_scenario_result() -> ScenarioResult:
    scenario_ = Mock(VirtualScenario)
    return ScenarioResult(scenario_)


@pytest.mark.asyncio
async def test_terminator_plugin_passed(*, terminator: TerminatorPlugin, dispatcher: Dispatcher):
    with given:
        report = Report()
        scenario_result = make_scenario_result().mark_passed()
        report.add_result(scenario_result)

    with when, raises(BaseException) as exception:
        await dispatcher.fire(CleanupEvent(report))

    with then:
        assert exception.type is SystemExit
        assert cast(SystemExit, exception.value).code == 0


@pytest.mark.asyncio
async def test_terminator_plugin_failed(*, terminator: TerminatorPlugin, dispatcher: Dispatcher):
    with given:
        report = Report()
        scenario_result = make_scenario_result().mark_failed()
        report.add_result(scenario_result)

    with when, raises(BaseException) as exception:
        await dispatcher.fire(CleanupEvent(report))

    with then:
        assert exception.type is SystemExit
        assert cast(SystemExit, exception.value).code == 1


@pytest.mark.asyncio
async def test_terminator_plugin_no_passed(*, terminator: TerminatorPlugin,
                                           dispatcher: Dispatcher):
    with given:
        report = Report()

    with when, raises(BaseException) as exception:
        await dispatcher.fire(CleanupEvent(report))

    with then:
        assert exception.type is SystemExit
        assert cast(SystemExit, exception.value).code == 1



import os
from unittest.mock import patch

from baby_steps import given, then, when

from vedro.plugins.director.rich.utils import get_terminal_size


def test_get_terminal_size():
    with given:
        width, height = 1, 2
        terminal_size = os.terminal_size((width, height))

    with when, patch("shutil.get_terminal_size", return_value=terminal_size):
        res = get_terminal_size()

    with then:
        assert res == terminal_size


def test_get_terminal_size_default_fallback():
    with given:
        width, height = 80, 24
        terminal_size = os.terminal_size((0, 0))

    with when, patch("shutil.get_terminal_size", return_value=terminal_size):
        res = get_terminal_size(width, height)

    with then:
        assert res == os.terminal_size((80, 24))


def test_get_terminal_size_custom_fallback():
    with given:
        width, height = 1, 2
        terminal_size = os.terminal_size((0, 0))

    with when, patch("shutil.get_terminal_size", return_value=terminal_size):
        res = get_terminal_size(width, height)

    with then:
        assert res == os.terminal_size((width, height))



from argparse import ArgumentParser, Namespace

import pytest
from baby_steps import given, then, when
from pytest import raises

from vedro.core import Dispatcher
from vedro.events import ArgParsedEvent, ArgParseEvent
from vedro.plugins.slicer import Slicer, SlicerPlugin


@pytest.fixture()
def dispatcher() -> Dispatcher:
    return Dispatcher()


@pytest.fixture()
def slicer(dispatcher: Dispatcher) -> SlicerPlugin:
    slicer = SlicerPlugin(Slicer)
    slicer.subscribe(dispatcher)
    return slicer


@pytest.mark.asyncio
async def test_slicer_plugin(*, slicer: SlicerPlugin, dispatcher: Dispatcher):
    with given:
        event = ArgParseEvent(ArgumentParser())

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None


@pytest.mark.parametrize(("slicer_total", "slicer_index"), [
    (None, None),
    (1, 0),
])
@pytest.mark.asyncio
async def test_slicer_plugin_arg_validation(slicer_total, slicer_index, *,
                                            slicer: SlicerPlugin,
                                            dispatcher: Dispatcher):
    with given:
        event = ArgParsedEvent(Namespace(slicer_total=slicer_total, slicer_index=slicer_index))

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None


@pytest.mark.parametrize(("slicer_total", "slicer_index"), [
    (1, None),  # slicer_index is None
    (None, 1),  # slicer_total is None
    (0, 1),  # slicer_total <= 0
    (1, -1),  # slicer_index < 0
    (1, 1),  # slicer_index > slicer_total
])
@pytest.mark.asyncio
async def test_slicer_plugin_arg_validation_error(slicer_total, slicer_index, *,
                                                  slicer: SlicerPlugin,
                                                  dispatcher: Dispatcher):
    with given:
        event = ArgParsedEvent(Namespace(slicer_total=slicer_total, slicer_index=slicer_index))

    with when, raises(Exception) as exc_info:
        await dispatcher.fire(event)

    with then:
        assert exc_info.type is AssertionError



from ._arg_parser import ArgumentParser
from ._artifacts import Artifact, FileArtifact, MemoryArtifact
from ._config_loader import Config, ConfigFileLoader, ConfigLoader, ConfigType, Section
from ._dispatcher import Dispatcher, Subscriber
from ._event import Event
from ._exc_info import ExcInfo
from ._lifecycle import Lifecycle
from ._module_loader import ModuleFileLoader, ModuleLoader
from ._plugin import Plugin, PluginConfig
from ._report import Report
from ._runner import Runner
from ._scenario_discoverer import ScenarioDiscoverer
from ._scenario_finder import ScenarioFinder
from ._scenario_loader import ScenarioLoader
from ._scenario_result import ScenarioResult, ScenarioStatus
from ._step_result import StepResult, StepStatus
from ._virtual_scenario import VirtualScenario
from ._virtual_step import VirtualStep

__all__ = ("Dispatcher", "Subscriber", "Event", "ExcInfo", "Lifecycle", "Plugin", "PluginConfig",
           "Report", "Runner", "ScenarioDiscoverer", "ScenarioFinder", "ScenarioLoader",
           "ScenarioResult", "StepResult", "VirtualScenario", "VirtualStep",
           "ScenarioStatus", "StepStatus", "ArgumentParser",
           "ConfigLoader", "ConfigFileLoader", "Config", "Section", "ConfigType",
           "ModuleLoader", "ModuleFileLoader", "Artifact", "MemoryArtifact", "FileArtifact",)



from argparse import Namespace
from pathlib import Path
from typing import List

from ..core._arg_parser import ArgumentParser
from ..core._config_loader import ConfigType
from ..core._event import Event
from ..core._exc_info import ExcInfo
from ..core._report import Report
from ..core._scenario_result import ScenarioResult
from ..core._step_result import StepResult
from ..core._virtual_scenario import VirtualScenario


class ConfigLoadedEvent(Event):
    def __init__(self, path: Path, config: ConfigType) -> None:
        self._path = path
        self._config = config

    @property
    def path(self) -> Path:
        return self._path

    @property
    def config(self) -> ConfigType:
        return self._config

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._path!r}, <Config>)"


class ArgParseEvent(Event):
    def __init__(self, arg_parser: ArgumentParser) -> None:
        self._arg_parser = arg_parser

    @property
    def arg_parser(self) -> ArgumentParser:
        return self._arg_parser

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._arg_parser!r})"


class ArgParsedEvent(Event):
    def __init__(self, args: Namespace) -> None:
        self._args = args

    @property
    def args(self) -> Namespace:
        return self._args

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._args!r})"


class StartupEvent(Event):
    def __init__(self, scenarios: List[VirtualScenario]) -> None:
        self._scenarios = scenarios

    @property
    def scenarios(self) -> List[VirtualScenario]:
        return self._scenarios

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._scenarios!r})"


class _ScenarioEvent(Event):
    def __init__(self, scenario_result: ScenarioResult) -> None:
        self._scenario_result = scenario_result

    @property
    def scenario_result(self) -> ScenarioResult:
        return self._scenario_result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._scenario_result!r})"


class ScenarioSkippedEvent(_ScenarioEvent):
    pass


class ScenarioFailedEvent(_ScenarioEvent):
    pass


class ScenarioRunEvent(_ScenarioEvent):
    pass


class ScenarioPassedEvent(_ScenarioEvent):
    pass


class _StepEvent(Event):
    def __init__(self, step_result: StepResult) -> None:
        self._step_result = step_result

    @property
    def step_result(self) -> StepResult:
        return self._step_result

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._step_result!r})"


class StepRunEvent(_StepEvent):
    pass


class StepFailedEvent(_StepEvent):
    pass


class StepPassedEvent(_StepEvent):
    pass


class ExceptionRaisedEvent(Event):
    def __init__(self, exc_info: ExcInfo) -> None:
        self._exc_info = exc_info

    @property
    def exc_info(self) -> ExcInfo:
        return self._exc_info

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._exc_info!r})"


class CleanupEvent(Event):
    def __init__(self, report: Report) -> None:
        self._report = report

    @property
    def report(self) -> Report:
        return self._report

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._report!r})"


__all__ = ("Event", "ConfigLoadedEvent", "ArgParseEvent", "ArgParsedEvent",
           "StartupEvent", "ScenarioRunEvent", "ScenarioSkippedEvent",
           "ScenarioFailedEvent", "ScenarioPassedEvent",
           "StepRunEvent", "StepFailedEvent", "StepPassedEvent", "ExceptionRaisedEvent",
           "CleanupEvent")



from typing import Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import ArgParsedEvent, ArgParseEvent, StartupEvent

__all__ = ("Slicer", "SlicerPlugin",)


class SlicerPlugin(Plugin):
    def __init__(self, config: Type["Slicer"]) -> None:
        super().__init__(config)
        self._total: Union[int, None] = None
        self._index: Union[int, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Slicer")
        group.add_argument("--slicer-total", type=int, help="Set total workers")
        group.add_argument("--slicer-index", type=int, help="Set current worker")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._total = event.args.slicer_total
        self._index = event.args.slicer_index
        if self._total is not None:
            assert self._index is not None
            assert self._total > 0
        if self._index is not None:
            assert self._total is not None
            assert 0 <= self._index < self._total

    def on_startup(self, event: StartupEvent) -> None:
        if (self._total is None) or (self._index is None):
            return
        index = 0
        for scenario in event.scenarios:
            if scenario.is_skipped():
                continue
            if index % self._total != self._index:
                scenario.skip()
            index += 1


class Slicer(PluginConfig):
    plugin = SlicerPlugin



import json
import os
from traceback import format_exception
from types import TracebackType
from typing import Any, Callable, Dict, Generator, List, Tuple, Type, Union

from rich.console import Console
from rich.status import Status
from rich.style import Style
from rich.traceback import Traceback

import vedro
from vedro.core import Dispatcher, PluginConfig, ScenarioResult, StepResult
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)

from .._director_init_event import DirectorInitEvent
from .._reporter import Reporter
from .utils import make_console

__all__ = ("RichReporterPlugin", "RichReporter",)


ScenarioEndEventType = Union[ScenarioPassedEvent, ScenarioFailedEvent, ScenarioSkippedEvent]


class RichReporterPlugin(Reporter):
    def __init__(self, config: Type["RichReporter"], *,
                 console_factory: Callable[[], Console] = make_console) -> None:
        super().__init__(config)
        self._console = console_factory()
        self._verbosity = 0
        self._tb_pretty = config.tb_pretty
        self._tb_show_internal_calls = config.tb_show_internal_calls
        self._tb_show_locals = config.tb_show_locals
        self._tb_max_frames = config.tb_max_frames
        self._show_timings = config.show_timings
        self._show_paths = config.show_paths
        self._show_scenario_spinner = config.show_scenario_spinner
        self._scenario_spinner: Union[Status, None] = None
        self._namespace: Union[str, None] = None
        self._buffer: List[ScenarioResult] = []
        self._reruns = 0

    def subscribe(self, dispatcher: Dispatcher) -> None:
        super().subscribe(dispatcher)
        dispatcher.listen(DirectorInitEvent, lambda e: e.director.register("rich", self))

    def on_chosen(self) -> None:
        assert isinstance(self._dispatcher, Dispatcher)
        self._dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                        .listen(ArgParsedEvent, self.on_arg_parsed) \
                        .listen(StartupEvent, self.on_startup) \
                        .listen(ScenarioRunEvent, self.on_scenario_run) \
                        .listen(ScenarioSkippedEvent, self.on_scenario_end) \
                        .listen(ScenarioPassedEvent, self.on_scenario_end) \
                        .listen(ScenarioFailedEvent, self.on_scenario_end) \
                        .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Rich Reporter")

        help_message = ("Increase verbosity level "
                        "(-v: show steps, -vv: show exception, -vvv: show scope)")
        group.add_argument("-v", "--verbose",
                           action="count",
                           default=self._verbosity,
                           help=help_message)
        group.add_argument("--show-timings",
                           action="store_true",
                           default=self._show_timings,
                           help="Show the elapsed time of each scenario")
        group.add_argument("--show-paths",
                           action="store_true",
                           default=self._show_paths,
                           help="Show the relative path of each passed scenario")
        group.add_argument("--tb-show-internal-calls",
                           action="store_true",
                           default=self._tb_show_internal_calls,
                           help="Show internal calls in the traceback output")
        group.add_argument("--tb-show-locals",
                           action="store_true",
                           default=self._tb_show_locals,
                           help="Show local variables in the traceback output")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._verbosity = event.args.verbose
        self._show_timings = event.args.show_timings
        self._show_paths = event.args.show_paths
        self._tb_show_internal_calls = event.args.tb_show_internal_calls
        self._tb_show_locals = event.args.tb_show_locals
        self._reruns = event.args.reruns

    def on_startup(self, event: StartupEvent) -> None:
        self._console.out("Scenarios")

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        if event.scenario_result.scenario.namespace != self._namespace:
            self._namespace = event.scenario_result.scenario.namespace
            namespace = self._namespace.replace("_", " ").replace("/", " / ")
            self._console.out(f"* {namespace}", style=Style(bold=True))
        if self._show_scenario_spinner:
            self._scenario_spinner = self._console.status(event.scenario_result.scenario.subject)
            self._scenario_spinner.start()

    def on_scenario_end(self, event: ScenarioEndEventType) -> None:
        if self._scenario_spinner:
            self._scenario_spinner.stop()
            self._scenario_spinner = None

        if not event.scenario_result.is_failed() and len(self._buffer) == 0:
            self._buffer.append(event.scenario_result)
            self._print_buffered()
        else:
            self._buffer.append(event.scenario_result)
            if event.scenario_result.rerun == self._reruns:
                self._print_buffered()

    def _print_scenario_skipped(self, scenario_result: ScenarioResult, *, indent: int = 0) -> None:
        pass

    def _print_scenario_passed(self, scenario_result: ScenarioResult, *, indent: int = 0) -> None:
        self._print_scenario_subject(scenario_result, self._show_timings)
        if self._show_paths:
            prepend = " " * indent
            rel_path = scenario_result.scenario.path.relative_to(os.getcwd())
            self._console.out(f"{prepend}   > {rel_path}", style=Style(color="grey50"))

    def _print_scenario_failed(self, scenario_result: ScenarioResult, *, indent: int = 0) -> None:
        self._print_scenario_subject(scenario_result, self._show_timings)

        if self._verbosity > 0:
            for step_result in scenario_result.step_results:
                self._print_step_name(step_result, indent=4 + indent)

        if self._verbosity > 1:
            for step_result in scenario_result.step_results:
                if step_result.exc_info:
                    self._print_exception(step_result.exc_info.value,
                                          step_result.exc_info.traceback)

        if self._verbosity > 2:
            if scenario_result.scope:
                self._print_scope(scenario_result.scope)
                self._console.out(" ")

    def _find_resolution(self, scenario_results: List[ScenarioResult]) -> ScenarioResult:
        passed, failed = [], []
        for scenario_result in scenario_results:
            if scenario_result.is_passed():
                passed.append(scenario_result)
            else:
                failed.append(scenario_result)
        return passed[-1] if len(passed) > len(failed) else failed[-1]

    def _print_scenario_result(self, scenario_result: ScenarioResult, *, indent: int = 0) -> None:
        if scenario_result.is_passed():
            self._print_scenario_passed(scenario_result, indent=indent)
        elif scenario_result.is_failed():
            self._print_scenario_failed(scenario_result, indent=indent)
        else:
            self._print_scenario_skipped(scenario_result, indent=indent)

    def _print_buffered(self) -> None:
        if len(self._buffer) == 1:
            return self._print_scenario_result(self._buffer.pop())

        resolution = self._find_resolution(self._buffer)
        self._print_scenario_subject(resolution)

        for rerun in range(1, len(self._buffer) + 1):
            scenario_result = self._buffer.pop(0)
            prefix = f" ├─[{rerun}/{self._reruns + 1}]"
            self._console.out(" │")
            self._console.out(prefix, end="")
            self._print_scenario_result(scenario_result, indent=len(prefix))

        self._console.out(" ")

    def _print_scenario_subject(self, scenario_result: ScenarioResult,
                                show_timings: bool = False) -> None:
        if scenario_result.is_passed():
            subject = f" ✔ {scenario_result.scenario.subject}"
            style = Style(color="green")
        elif scenario_result.is_failed():
            subject = f" ✗ {scenario_result.scenario.subject}"
            style = Style(color="red")
        else:
            return

        if show_timings:
            self._console.out(subject, style=style, end="")
            self._console.out(f" ({scenario_result.elapsed:.2f}s)", style=Style(color="grey50"))
        else:
            self._console.out(subject, style=style)

    def _print_step_name(self, step_result: StepResult, *, indent: int = 0) -> None:
        prepend = " " * indent
        if step_result.is_passed():
            name = f"{prepend}✔ {step_result.step_name}"
            style = Style(color="green")
        elif step_result.is_failed():
            name = f"{prepend}✗ {step_result.step_name}"
            style = Style(color="red")
        else:
            return

        if self._show_timings:
            self._console.out(name, style=style, end="")
            self._console.out(f" ({step_result.elapsed:.2f}s)", style=Style(color="grey50"))
        else:
            self._console.out(name, style=style)

    def _filter_locals(self, local_variables: Dict[str, Any]) -> Dict[str, Any]:
        filtered_locals = {}
        for key, val in local_variables.items():
            # self is always "Scenario"
            if key == "self":
                continue
            # assertion rewriter stuff
            if key.startswith("@"):
                continue
            filtered_locals[key] = val
        return filtered_locals

    def _print_exception(self, exception: BaseException, traceback: TracebackType) -> None:
        if not self._tb_show_internal_calls:
            root = os.path.dirname(vedro.__file__)
            while traceback.tb_next is not None:
                filename = os.path.abspath(traceback.tb_frame.f_code.co_filename)
                if os.path.commonpath([root, filename]) != root:
                    break
                traceback = traceback.tb_next

        if self._tb_pretty:
            trace = Traceback.extract(type(exception), exception, traceback,
                                      show_locals=self._tb_show_locals)
            for stack in trace.stacks:
                for frame in stack.frames:
                    if frame.locals:
                        frame.locals = self._filter_locals(frame.locals)
            self._console.print(Traceback(trace,
                                          max_frames=self._tb_max_frames, word_wrap=True))
            self._console.out(" ")
        else:
            formatted = format_exception(type(exception), exception, traceback,
                                         limit=self._tb_max_frames)
            self._console.out("".join(formatted), style=Style(color="yellow"))

    def _format_scope(self, scope: Dict[Any, Any]) -> Generator[Tuple[str, str], None, None]:
        for key, val in scope.items():
            try:
                val_repr = json.dumps(val, ensure_ascii=False, indent=4)
            except:  # noqa: E722
                val_repr = repr(val)
            yield str(key), val_repr

    def _print_scope(self, scope: Dict[Any, Any]) -> None:
        self._console.out("Scope:", style=Style(color="blue", bold=True))
        for key, val in self._format_scope(scope):
            self._console.out(f" {key}: ", end="", style=Style(color="blue"))
            self._console.out(val)

    def on_cleanup(self, event: CleanupEvent) -> None:
        if event.report.failed == 0 and event.report.passed > 0:
            style = Style(color="green", bold=True)
        else:
            style = Style(color="red", bold=True)
        self._console.out(" ")

        if len(event.report.summary) > 0:
            summary = "# " + "\n# ".join(event.report.summary)
            self._console.out(summary, style=Style(color="grey70"))

        if event.report.total == 1:
            scenario = "scenario"
        else:
            scenario = "scenarios"
        self._console.out(f"# {event.report.total} {scenario}, "
                          f"{event.report.passed} passed, "
                          f"{event.report.failed} failed, "
                          f"{event.report.skipped} skipped",
                          style=style,
                          end="")
        self._console.out(f" ({event.report.elapsed:.2f}s)", style=Style(color="blue"))


class RichReporter(PluginConfig):
    plugin = RichReporterPlugin

    # Show the elapsed time of each scenario
    show_timings: bool = False

    # Show the relative path of each passed scenario
    show_paths: bool = False

    # Show status indicator of the current running scenario
    show_scenario_spinner: bool = False

    # Show pretty traceback
    tb_pretty: bool = True

    # Show internal calls in the traceback output
    tb_show_internal_calls: bool = False

    # Show local variables in the traceback output
    tb_show_locals: bool = False

    # Max stack trace entries to show
    tb_max_frames: int = 8



import os
from pathlib import Path
from typing import AsyncGenerator, Optional

from ._file_filters import FileFilter
from ._scenario_finder import ScenarioFinder

__all__ = ("ScenarioFileFinder",)


class ScenarioFileFinder(ScenarioFinder):
    def __init__(self, file_filter: Optional[FileFilter] = None,
                 dir_filter: Optional[FileFilter] = None) -> None:
        self._file_filter = file_filter
        self._dir_filter = dir_filter

    async def find(self, root: Path) -> AsyncGenerator[Path, None]:
        for path, dirs, files in os.walk(root):
            for dirname in dirs:
                dir_path = Path(os.path.join(path, dirname))
                if self._dir_filter and self._dir_filter.filter(dir_path):
                    dirs.remove(dirname)
            for filename in files:
                file_path = Path(os.path.join(path, filename))
                if self._file_filter and self._file_filter.filter(file_path):
                    continue
                yield file_path



import random
from typing import Any, Dict, Type, Union, cast
from uuid import uuid4

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioRunEvent,
    StartupEvent,
)

__all__ = ("Seeder", "SeederPlugin",)


class SeederPlugin(Plugin):
    def __init__(self, config: Type["Seeder"], *, random: Any = random) -> None:
        super().__init__(config)
        self._random = random
        self._seed: Union[str, None] = None
        self._scenarios: Dict[str, int] = {}

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("--seed", nargs="?", help="Set seed")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._seed = event.args.seed
        if self._seed is None:
            self._seed = str(uuid4())

    def _generate_seed(self) -> int:
        return cast(int, self._random.randint(1, 2 ** 63 - 1))

    def on_startup(self, event: StartupEvent) -> None:
        self._random.seed(self._seed)
        for scenario in event.scenarios:
            self._scenarios[scenario.unique_id] = self._generate_seed()

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        seed = self._scenarios[event.scenario_result.scenario.unique_id]
        self._random.seed(seed)
        for _ in range(event.scenario_result.rerun):
            seed = self._generate_seed()
        self._random.seed(seed)

    def on_cleanup(self, event: CleanupEvent) -> None:
        if (event.report.passed + event.report.failed) > 0:
            summary = f"--seed {self._seed}"
            event.report.add_summary(summary)


class Seeder(PluginConfig):
    plugin = SeederPlugin



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from collections import deque
from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when

from vedro.core import Dispatcher, Plugin, ScenarioResult, VirtualScenario
from vedro.events import ScenarioFailedEvent, ScenarioPassedEvent, ScenarioRunEvent
from vedro.plugins.deferrer import Deferrer, DeferrerPlugin


@pytest.fixture()
def dispatcher() -> Dispatcher:
    return Dispatcher()


@pytest.fixture()
def queue() -> deque:
    return deque()


@pytest.fixture()
def deferrer(queue) -> DeferrerPlugin:
    return DeferrerPlugin(Deferrer, queue=queue)


def test_deferrer_plugin():
    with when:
        plugin = DeferrerPlugin(Deferrer)

    with then:
        assert isinstance(plugin, Plugin)


@pytest.mark.asyncio
async def test_deferrer_scenario_run_event(*, dispatcher: Dispatcher,
                                           deferrer: DeferrerPlugin, queue: deque):
    with given:
        deferrer.subscribe(dispatcher)

        scenario_result = ScenarioResult(Mock(VirtualScenario))
        event = ScenarioRunEvent(scenario_result)

        queue.append((Mock(), (), {}))

    with when:
        await dispatcher.fire(event)

    with then:
        assert len(queue) == 0


@pytest.mark.asyncio
@pytest.mark.parametrize("event_class", [ScenarioPassedEvent, ScenarioFailedEvent])
async def test_deferrer_scenario_end_event(event_class, *, dispatcher: Dispatcher,
                                           deferrer: DeferrerPlugin, queue: deque):
    with given:
        deferrer.subscribe(dispatcher)
        scenario_result = ScenarioResult(Mock(VirtualScenario))
        event = event_class(scenario_result)

        manager = Mock()
        deferred1 = Mock()
        manager.attach_mock(deferred1, "deferred1")
        deferred2 = Mock()
        manager.attach_mock(deferred2, "deferred2")

        args1, kwargs1 = ("arg1", "arg2"), {"key1": "val1"}
        queue.append((deferred1, args1, kwargs1))
        args2, kwargs2 = ("arg3", "arg4"), {"key2": "val2"}
        queue.append((deferred2, args2, kwargs2))

    with when:
        await dispatcher.fire(event)

    with then:
        assert manager.mock_calls == [
            call.deferred2(*args2, **kwargs2),
            call.deferred1(*args1, **kwargs1),
        ]
        assert len(queue) == 0


@pytest.mark.asyncio
@pytest.mark.parametrize("event_class", [ScenarioPassedEvent, ScenarioFailedEvent])
async def test_deferrer_scenario_end_event_async(event_class, *, dispatcher: Dispatcher,
                                                 deferrer: DeferrerPlugin, queue: deque):
    with given:
        deferrer.subscribe(dispatcher)
        scenario_result = ScenarioResult(Mock(VirtualScenario))
        event = event_class(scenario_result)

        manager = Mock()
        deferred1 = AsyncMock()
        manager.attach_mock(deferred1, "deferred1")
        deferred2 = AsyncMock()
        manager.attach_mock(deferred2, "deferred2")

        args1, kwargs1 = ("arg1", "arg2"), {"key1": "val1"}
        queue.append((deferred1, args1, kwargs1))
        args2, kwargs2 = ("arg3", "arg4"), {"key2": "val2"}
        queue.append((deferred2, args2, kwargs2))

    with when:
        await dispatcher.fire(event)

    with then:
        assert manager.mock_calls == [
            call.deferred2(*args2, **kwargs2),
            call.deferred1(*args1, **kwargs1),
        ]
        deferred1.assert_awaited_once()
        deferred2.assert_awaited_once()
        assert len(queue) == 0



from pathlib import Path

import pytest
from baby_steps import given, then, when

from vedro.core import Config, ConfigFileLoader


@pytest.mark.asyncio
async def test_config_file_loader(*, tmp_path: Path):
    with given:
        config = Config
        loader = ConfigFileLoader(config)
        path = tmp_path / "vedro.cfg.py"
        path.write_text("\n".join([
            "import vedro",
            "class Config(vedro.Config):",
            "    pass",
        ]))

    with when:
        res = await loader.load(path)

    with then:
        assert res != config


@pytest.mark.asyncio
async def test_config_file_loader_empty_file(*, tmp_path: Path):
    with given:
        config = Config
        loader = ConfigFileLoader(config)
        path = tmp_path / "vedro.cfg.py"
        path.write_text("")

    with when:
        res = await loader.load(path)

    with then:
        assert res == config


@pytest.mark.asyncio
async def test_config_file_loader_not_exist(*, tmp_path: Path):
    with given:
        config = Config
        loader = ConfigFileLoader(config)
        path = tmp_path / "vedro.cfg.py"

    with when:
        res = await loader.load(path)

    with then:
        assert res == config
